from benchling_api_client.v2.beta.api.custom_entities import bulk_upsert_custom_entities, upsert_custom_entity
from benchling_api_client.v2.beta.models.custom_entities_bulk_upsert_request import (
    CustomEntitiesBulkUpsertRequest,
)
from benchling_api_client.v2.beta.models.custom_entity_upsert_request import CustomEntityUpsertRequest

from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.models import AsyncTaskLink, CustomEntity
from benchling_sdk.services.v2.base_service import BaseService


class V2BetaCustomEntityService(BaseService):
    """
    V2-Beta Custom Entities.

    Benchling supports custom entities for biological entities that are neither sequences or proteins. Custom
    entities must have an entity schema set and can have both schema fields and custom fields.

    See https://benchling.com/api/v2-beta/reference#/Custom%20Entities
    """

    @api_method
    def upsert(self, entity_registry_id: str, entity: CustomEntityUpsertRequest) -> CustomEntity:
        """
        Create or modify a custom entity.

        See https://benchling.com/api/v2-beta/reference#/Custom%20Entities/upsertCustomEntity
        """
        response = upsert_custom_entity.sync_detailed(
            client=self.client, entity_registry_id=entity_registry_id, json_body=entity
        )
        return model_from_detailed(response)

    @api_method
    def bulk_upsert(self, body: CustomEntitiesBulkUpsertRequest) -> AsyncTaskLink:
        """
        Bulk update custom entities.

        See https://benchling.com/api/v2-beta/reference#/Custom%20Entities/bulkUpsertCustomEntities
        """
        response = bulk_upsert_custom_entities.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)



from typing import List, Optional, Union

from benchling_api_client.v2.stable.api.dna_alignments import (
    create_dna_consensus_alignment,
    create_dna_template_alignment,
    delete_dna_alignment,
    get_dna_alignment,
    list_dna_alignments,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.constants import _translate_to_string_enum
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.logging_helpers import log_deprecation
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset, optional_array_query_param
from benchling_sdk.models import (
    AsyncTaskLink,
    DnaAlignment,
    DnaAlignmentsPaginatedList,
    DnaAlignmentSummary,
    DnaConsensusAlignmentCreate,
    DnaTemplateAlignmentCreate,
    ListDNAAlignmentsSort,
)
from benchling_sdk.services.v2.base_service import BaseService


class DnaAlignmentsService(BaseService):
    """
    DNA Alignments (Deprecated).

    A DNA alignment is a Benchling object representing an alignment of multiple DNA sequences.

    Please migrate to the corresponding Nucleotide Alignment endpoints.

    See https://benchling.com/api/reference#/DNA%20Alignments
    See https://benchling.com/api/reference#/Nucleotide%20Alignments
    """

    @api_method
    def get_by_id(self, dna_alignment_id: str) -> DnaAlignment:
        """
        Get a DNA alignment (Deprecated).

        Please migrate to the corresponding Nucleotide Alignment endpoints.

        See https://benchling.com/api/reference#/DNA%20Alignments/getDNAAlignment
        See https://benchling.com/api/reference#/Nucleotide%20Alignments/getNucleotideAlignment
        """
        log_deprecation("dna_alignments.get_by_id", "nucleotide_alignments.get_by_id")
        response = get_dna_alignment.sync_detailed(client=self.client, dna_alignment_id=dna_alignment_id)
        return model_from_detailed(response)

    @api_method
    def _dna_alignments_page(
        self,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        ids: Optional[List[str]] = None,
        names_any_of: Optional[List[str]] = None,
        names_any_of_case_sensitive: Optional[List[str]] = None,
        sequence_ids: Optional[List[str]] = None,
        sort: Optional[ListDNAAlignmentsSort] = None,
        page_size: Optional[int] = None,
        next_token: NextToken = None,
    ) -> Response[DnaAlignmentsPaginatedList]:
        response = list_dna_alignments.sync_detailed(
            client=self.client,
            modified_at=none_as_unset(modified_at),
            name=none_as_unset(name),
            name_includes=none_as_unset(name_includes),
            ids=none_as_unset(optional_array_query_param(ids)),
            namesany_of=none_as_unset(optional_array_query_param(names_any_of)),
            namesany_ofcase_sensitive=none_as_unset(optional_array_query_param(names_any_of_case_sensitive)),
            sequence_ids=none_as_unset(optional_array_query_param(sequence_ids)),
            sort=none_as_unset(sort),
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(
        self,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        ids: Optional[List[str]] = None,
        names_any_of: Optional[List[str]] = None,
        names_any_of_case_sensitive: Optional[List[str]] = None,
        sequence_ids: Optional[List[str]] = None,
        sort: Optional[Union[str, ListDNAAlignmentsSort]] = None,
        page_size: Optional[int] = None,
    ) -> PageIterator[DnaAlignmentSummary]:
        """
        List DNA Alignments (Deprecated).

        Please migrate to the corresponding Nucleotide Alignment endpoints.

        See https://benchling.com/api/reference#/DNA%20Alignments/listDNAAlignments
        See https://benchling.com/api/reference#/Nucleotide%20Alignments/listNucleotideAlignments
        """
        log_deprecation("dna_alignments.list", "nucleotide_alignments.list")

        def api_call(next_token: NextToken) -> Response[DnaAlignmentsPaginatedList]:
            return self._dna_alignments_page(
                modified_at=modified_at,
                name=name,
                name_includes=name_includes,
                ids=ids,
                names_any_of=names_any_of,
                names_any_of_case_sensitive=names_any_of_case_sensitive,
                sequence_ids=sequence_ids,
                sort=_translate_to_string_enum(ListDNAAlignmentsSort, sort),
                page_size=page_size,
                next_token=next_token,
            )

        def results_extractor(body: DnaAlignmentsPaginatedList) -> Optional[List[DnaAlignmentSummary]]:
            return body.dna_alignments

        return PageIterator(api_call, results_extractor)

    @api_method
    def create_template_alignment(self, template_alignment: DnaTemplateAlignmentCreate) -> AsyncTaskLink:
        """
        Create a template DNA alignment (Deprecated).

        Please migrate to the corresponding Nucleotide Alignment endpoints.

        See https://benchling.com/api/reference#/DNA%20Alignments/createTemplateAlignment
        See https://benchling.com/api/reference#/Nucleotide%20Alignments/createTemplateNucleotideAlignment
        """
        log_deprecation(
            "dna_alignments.create_template_alignment", "nucleotide_alignments.create_template_alignment"
        )
        response = create_dna_template_alignment.sync_detailed(
            client=self.client, json_body=template_alignment
        )
        return model_from_detailed(response)

    @api_method
    def create_consensus_alignment(self, consensus_alignment: DnaConsensusAlignmentCreate) -> AsyncTaskLink:
        """
        Create a consensus DNA alignment (Deprecated).

        Please migrate to the corresponding Nucleotide Alignment endpoints.

        See https://benchling.com/api/reference#/DNA%20Alignments/createConsensusAlignment
        See https://benchling.com/api/reference#/Nucleotide%20Alignments/createConsensusNucleotideAlignment
        """
        log_deprecation(
            "dna_alignments.create_consensus_alignment", "nucleotide_alignments.create_consensus_alignment"
        )
        response = create_dna_consensus_alignment.sync_detailed(
            client=self.client, json_body=consensus_alignment
        )
        return model_from_detailed(response)

    @api_method
    def delete_alignment(self, dna_alignment_id: str) -> None:
        """
        Delete a DNA alignment (Deprecated).

        Please migrate to the corresponding Nucleotide Alignment endpoints.

        See https://benchling.com/api/reference#/DNA%20Alignments/deleteDNAAlignment
        See https://benchling.com/api/reference#/Nucleotide%20Alignments/deleteNucleotideAlignment
        """
        log_deprecation("dna_alignments.delete_alignment", "nucleotide_alignments.delete_alignment")
        response = delete_dna_alignment.sync_detailed(client=self.client, dna_alignment_id=dna_alignment_id)
        raise_for_status(response)



from typing import Iterable, List, Optional, Union

from benchling_api_client.v2.beta.api.apps import (
    archive_canvases,
    create_canvas,
    create_session,
    get_app_configuration_by_app_id,
    get_benchling_app_manifest,
    get_canvas,
    get_session_by_id,
    list_canvases,
    list_sessions,
    put_benchling_app_manifest,
    unarchive_canvases,
    update_canvas,
    update_session,
)
from benchling_api_client.v2.beta.models.benchling_app_configuration import BenchlingAppConfiguration
from benchling_api_client.v2.beta.models.benchling_app_manifest import BenchlingAppManifest
from benchling_api_client.v2.beta.models.canvas import Canvas
from benchling_api_client.v2.beta.models.canvas_create import CanvasCreate
from benchling_api_client.v2.beta.models.canvas_update import CanvasUpdate
from benchling_api_client.v2.beta.models.canvases_archival_change import CanvasesArchivalChange
from benchling_api_client.v2.beta.models.canvases_archive import CanvasesArchive
from benchling_api_client.v2.beta.models.canvases_archive_reason import CanvasesArchiveReason
from benchling_api_client.v2.beta.models.canvases_paginated_list import CanvasesPaginatedList
from benchling_api_client.v2.beta.models.canvases_unarchive import CanvasesUnarchive
from benchling_api_client.v2.beta.models.list_canvases_enabled import ListCanvasesEnabled
from benchling_api_client.v2.beta.models.list_canvases_sort import ListCanvasesSort
from benchling_api_client.v2.beta.models.session import Session
from benchling_api_client.v2.beta.models.session_create import SessionCreate
from benchling_api_client.v2.beta.models.session_update import SessionUpdate
from benchling_api_client.v2.beta.models.sessions_paginated_list import SessionsPaginatedList
from benchling_api_client.v2.stable.types import Response

from benchling_sdk.errors import AppSessionClosedError, raise_for_status
from benchling_sdk.helpers.constants import _translate_to_string_enum
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.logging_helpers import log_deprecation
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import none_as_unset
from benchling_sdk.services.v2.base_service import BaseService


class V2BetaAppService(BaseService):
    """
    V2-Beta Apps.

    Create and manage Apps on your tenant.

    https://benchling.com/api/v2-beta/reference?stability=not-available#/Apps
    """

    @api_method
    def get_configuration_by_app_id(self, app_id: str) -> BenchlingAppConfiguration:
        """
        Get an app's configuration by app id.

        See https://benchling.com/api/v2-beta/reference?stability=la#/Apps/getAppConfigurationByAppID

        Deprecated in favor of apps.list_app_configuration_items.
        """
        log_deprecation("v2.beta.apps.get_configuration_by_app_id", "apps.list_app_configuration_items")
        response = get_app_configuration_by_app_id.sync_detailed(client=self.client, app_id=app_id)
        return model_from_detailed(response)

    @api_method
    def get_manifest(self, app_id: str) -> BenchlingAppManifest:
        """
        Get app manifest.

        See https://benchling.com/api/v2-beta/reference#/Apps/getBenchlingAppManifest
        """
        response = get_benchling_app_manifest.sync_detailed(client=self.client, app_id=app_id)
        return model_from_detailed(response)

    @api_method
    def update_manifest(self, app_id: str, manifest: BenchlingAppManifest) -> BenchlingAppManifest:
        """
        Update an app manifest.

        See https://benchling.com/api/v2-beta/reference#/Apps/putBenchlingAppManifest
        """
        response = put_benchling_app_manifest.sync_detailed(
            client=self.client, app_id=app_id, yaml_body=manifest
        )
        return model_from_detailed(response)

    @api_method
    def create_canvas(self, canvas: CanvasCreate) -> Canvas:
        """
        Create an App Canvas that a Benchling App can write to and read user interaction from.

        See https://benchling.com/api/v2-beta/reference?stability=not-available#/Apps/createCanvas
        """
        response = create_canvas.sync_detailed(
            client=self.client,
            json_body=canvas,
        )
        return model_from_detailed(response)

    @api_method
    def _canvases_page(
        self,
        app_id: Optional[str] = None,
        feature_id: Optional[str] = None,
        resource_id: Optional[str] = None,
        enabled: Optional[ListCanvasesEnabled] = None,
        archive_reason: Optional[str] = None,
        sort: Optional[Union[str, ListCanvasesSort]] = None,
        next_token: NextToken = None,
        page_size: Optional[int] = None,
    ) -> Response[CanvasesPaginatedList]:
        response = list_canvases.sync_detailed(
            client=self.client,
            app_id=none_as_unset(app_id),
            feature_id=none_as_unset(feature_id),
            resource_id=none_as_unset(resource_id),
            enabled=none_as_unset(enabled),
            archive_reason=none_as_unset(archive_reason),
            sort=none_as_unset(_translate_to_string_enum(ListCanvasesSort, sort)),
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list_canvases(
        self,
        app_id: Optional[str] = None,
        feature_id: Optional[str] = None,
        resource_id: Optional[str] = None,
        enabled: Optional[ListCanvasesEnabled] = None,
        archive_reason: Optional[str] = None,
        sort: Optional[Union[str, ListCanvasesSort]] = None,
        page_size: Optional[int] = None,
    ) -> PageIterator[Session]:
        """
        List canvases.

        See https://benchling.com/api/v2-beta/reference?availability=not-available#/Apps/listCanvases
        """

        def api_call(next_token: NextToken) -> Response[CanvasesPaginatedList]:
            return self._canvases_page(
                app_id=app_id,
                feature_id=feature_id,
                resource_id=resource_id,
                enabled=enabled,
                archive_reason=archive_reason,
                sort=sort,
                next_token=next_token,
                page_size=page_size,
            )

        def results_extractor(body: CanvasesPaginatedList) -> Optional[List[Canvas]]:
            return body.canvases

        return PageIterator(api_call, results_extractor)

    @api_method
    def get_canvas(self, canvas_id: str) -> Canvas:
        """
        Get the current state of the App Canvas, including user input elements.

        See https://benchling.com/api/v2-beta/reference?stability=not-available#/Apps/getCanvas
        """
        response = get_canvas.sync_detailed(
            client=self.client,
            canvas_id=canvas_id,
        )
        return model_from_detailed(response)

    @api_method
    def update_canvas(self, canvas_id: str, canvas: CanvasUpdate) -> Canvas:
        """
        Update App Canvas.

        See https://benchling.com/api/v2-beta/reference?stability=not-available#/Apps/updateCanvas
        """
        response = update_canvas.sync_detailed(
            client=self.client,
            canvas_id=canvas_id,
            json_body=canvas,
        )
        return model_from_detailed(response)

    @api_method
    def archive_canvases(
        self, canvas_ids: Iterable[str], reason: CanvasesArchiveReason
    ) -> CanvasesArchivalChange:
        """
        Archive App Canvases.

        See https://benchling.com/api/v2-beta/reference?stability=not-available#/Apps/archiveCanvases
        """
        archive_request = CanvasesArchive(reason=reason, canvas_ids=list(canvas_ids))
        response = archive_canvases.sync_detailed(
            client=self.client,
            json_body=archive_request,
        )
        return model_from_detailed(response)

    @api_method
    def unarchive_canvases(self, canvas_ids: Iterable[str]) -> CanvasesArchivalChange:
        """
        Unarchive App Canvases.

        See https://benchling.com/api/v2-beta/reference?stability=not-available#/Apps/unarchiveCanvases
        """
        unarchive_request = CanvasesUnarchive(canvas_ids=list(canvas_ids))
        response = unarchive_canvases.sync_detailed(client=self.client, json_body=unarchive_request)
        return model_from_detailed(response)

    # Sessions

    @api_method
    def create_session(self, session: SessionCreate) -> Session:
        """
        Create a new session. Sessions cannot be archived once created.

        See https://benchling.com/api/v2-beta/reference?availability=not-available#/Apps/createSession
        """
        response = create_session.sync_detailed(
            client=self.client,
            json_body=session,
        )
        return model_from_detailed(response)

    @api_method
    def get_session_by_id(self, session_id: str) -> Session:
        """
        Get a session.

        See https://benchling.com/api/v2-beta/reference?availability=not-available#/Apps/getSessionById
        """
        response = get_session_by_id.sync_detailed(
            client=self.client,
            id=session_id,
        )
        return model_from_detailed(response)

    @api_method
    def update_session(self, session_id: str, session: SessionUpdate) -> Session:
        """
        Update session.

        Raises AppSessionClosedError if trying to update a Session that has already been closed.

        See https://benchling.com/api/v2-beta/reference?availability=not-available#/Apps/updateSession
        """
        response = update_session.sync_detailed(
            client=self.client,
            id=session_id,
            json_body=session,
        )
        return model_from_detailed(response, error_types=[AppSessionClosedError])

    @api_method
    def _sessions_page(
        self,
        app_id: Optional[str] = None,
        next_token: NextToken = None,
        page_size: Optional[int] = None,
    ) -> Response[SessionsPaginatedList]:
        response = list_sessions.sync_detailed(
            client=self.client,
            app_id=none_as_unset(app_id),
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list_sessions(
        self, app_id: Optional[str] = None, page_size: Optional[int] = None
    ) -> PageIterator[Session]:
        """
        List all sessions.

        See https://benchling.com/api/v2-beta/reference?availability=not-available#/Apps/listSessions
        """

        def api_call(next_token: NextToken) -> Response[SessionsPaginatedList]:
            return self._sessions_page(
                app_id=app_id,
                next_token=next_token,
                page_size=page_size,
            )

        def results_extractor(body: SessionsPaginatedList) -> Optional[List[Session]]:
            return body.sessions

        return PageIterator(api_call, results_extractor)



from enum import Enum
from typing import Optional, Type, TypeVar, Union

from benchling_api_client.v2.extensions import Enums


class CommonSortValues(str, Enum):
    """
    A list of commonly used sort parameters.

    These can be used as parameters to listing endpoints which can specify `sort=`.  If the chosen sort value is not
    supported by that endpoint, a bad request error (BenchlingError) will occur, with a message indicating so.
    """

    CREATEDAT = "createdAt"
    CREATEDATASC = "createdAt:asc"
    CREATEDATDESC = "createdAt:desc"
    MODIFIEDAT = "modifiedAt"
    MODIFIEDATASC = "modifiedAt:asc"
    MODIFIEDATDESC = "modifiedAt:desc"
    NAME = "name"
    NAMEASC = "name:asc"
    NAMEDESC = "name:desc"


_T = TypeVar("_T", bound=Enums.KnownString)


def _translate_to_string_enum(t: Type[_T], val: Union[None, _T, str]) -> Optional[_T]:
    if isinstance(val, t) or val is None:
        return val

    try:
        return t(val)
    except ValueError:
        strval = val.value if isinstance(val, Enum) else val
        return t.of_unknown(strval)  # type: ignore



from typing import Any, Dict, Iterable, List, Optional, Union

from benchling_api_client.v2.stable.api.plates import (
    archive_plates,
    bulk_get_plates,
    create_plate,
    get_plate,
    list_plates,
    unarchive_plates,
    update_plate,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.constants import _translate_to_string_enum
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.logging_helpers import check_for_csv_bug_fix
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import (
    none_as_unset,
    optional_array_query_param,
    schema_fields_query_param,
    unset_as_none,
)
from benchling_sdk.models import (
    ListPlatesSort,
    Plate,
    PlateCreate,
    PlatesArchivalChange,
    PlatesArchive,
    PlatesArchiveReason,
    PlatesPaginatedList,
    PlatesUnarchive,
    PlateUpdate,
)
from benchling_sdk.services.v2.base_service import BaseService

DEFAULT_PLATE_HTTP_TIMEOUT: float = 30.0


class PlateService(BaseService):
    """
    Plates.

    Plates are a structured storage type, grids of wells that each function like containers. Plates come in two
    types: a traditional "fixed" type, where the wells cannot move, and a "matrix" type. A matrix plate has similar
    functionality to a box, where the containers inside can be moved around and removed altogether.

    Plates are all associated with schemas, which define the type of the plate (e.g. "96 Well Plate") along with
    the fields that are tracked, the dimensions of the plate, and whether or not the plate is a matrix plate or a
    traditional well plate.

    Like all storage, every Plate has a barcode that is unique across the registry.

    See https://benchling.com/api/reference#/Plates
    """

    @api_method
    def get_by_id(self, plate_id: str, returning: Optional[Iterable[str]] = None) -> Plate:
        """
        Get a Plate by ID.

        See https://benchling.com/api/reference#/Plates/getPlate
        """
        returning_string = optional_array_query_param(returning)
        response = get_plate.sync_detailed(
            client=self.client, plate_id=plate_id, returning=none_as_unset(returning_string)
        )
        return model_from_detailed(response)

    @api_method
    def _plates_page(
        self,
        *,
        page_size: Optional[int] = 50,
        next_token: Optional[str] = None,
        sort: Optional[ListPlatesSort] = None,
        schema_id: Optional[str] = None,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        ancestor_storage_id: Optional[str] = None,
        storage_contents_id: Optional[str] = None,
        storage_contents_ids: Optional[List[str]] = None,
        ids: Optional[Iterable[str]] = None,
        barcodes: Optional[Iterable[str]] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        creator_ids: Optional[Iterable[str]] = None,
        archive_reason: Optional[str] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
        timeout_seconds: Optional[float] = None,
        returning: Optional[Iterable[str]] = None
    ) -> Response[PlatesPaginatedList]:
        timeout_client = self.client.with_timeout(timeout_seconds) if timeout_seconds else self.client
        response = list_plates.sync_detailed(
            client=timeout_client,
            sort=none_as_unset(sort),
            schema_id=none_as_unset(schema_id),
            modified_at=none_as_unset(modified_at),
            name=none_as_unset(name),
            name_includes=none_as_unset(name_includes),
            ancestor_storage_id=none_as_unset(ancestor_storage_id),
            storage_contents_id=none_as_unset(storage_contents_id),
            storage_contents_ids=none_as_unset(optional_array_query_param(storage_contents_ids)),
            archive_reason=none_as_unset(archive_reason),
            ids=none_as_unset(optional_array_query_param(ids)),
            barcodes=none_as_unset(optional_array_query_param(barcodes)),
            namesany_of=none_as_unset(optional_array_query_param(names_any_of)),
            namesany_ofcase_sensitive=none_as_unset(optional_array_query_param(names_any_of_case_sensitive)),
            creator_ids=none_as_unset(optional_array_query_param(creator_ids)),
            schema_fields=none_as_unset(schema_fields_query_param(schema_fields)),
            next_token=none_as_unset(next_token),
            page_size=none_as_unset(page_size),
            returning=none_as_unset(optional_array_query_param(returning)),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(
        self,
        *,
        sort: Optional[Union[str, ListPlatesSort]] = None,
        schema_id: Optional[str] = None,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        name_includes: Optional[str] = None,
        ancestor_storage_id: Optional[str] = None,
        storage_contents_id: Optional[str] = None,
        storage_contents_ids: Optional[List[str]] = None,
        ids: Optional[Iterable[str]] = None,
        barcodes: Optional[Iterable[str]] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        creator_ids: Optional[Iterable[str]] = None,
        archive_reason: Optional[str] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
        page_size: Optional[int] = None,
        timeout_seconds: Optional[float] = DEFAULT_PLATE_HTTP_TIMEOUT,
        returning: Optional[Iterable[str]] = None
    ) -> PageIterator[Plate]:
        """
        List Plates.

        List operations on large plates may take much longer than normal. The timeout_seconds
        parameter will use a higher HTTP timeout than the regular default. Pass a float to override
        it or pass None to use the standard client default.

        See https://benchling.com/api/reference#/Plates/listPlates
        """
        check_for_csv_bug_fix("storage_contents_ids", storage_contents_ids)
        if returning and "nextToken" not in returning:
            returning = list(returning) + ["nextToken"]

        def api_call(next_token: NextToken) -> Response[PlatesPaginatedList]:
            return self._plates_page(
                sort=_translate_to_string_enum(ListPlatesSort, sort),
                schema_id=schema_id,
                modified_at=modified_at,
                name=name,
                name_includes=name_includes,
                ancestor_storage_id=ancestor_storage_id,
                storage_contents_id=storage_contents_id,
                storage_contents_ids=storage_contents_ids,
                archive_reason=archive_reason,
                ids=ids,
                barcodes=barcodes,
                names_any_of=names_any_of,
                names_any_of_case_sensitive=names_any_of_case_sensitive,
                creator_ids=creator_ids,
                schema_fields=schema_fields,
                next_token=next_token,
                page_size=page_size,
                timeout_seconds=timeout_seconds,
                returning=returning,
            )

        def results_extractor(body: PlatesPaginatedList) -> Optional[List[Plate]]:
            return unset_as_none(body.plates)

        return PageIterator(api_call, results_extractor)

    @api_method
    def bulk_get(
        self,
        *,
        plate_ids: Optional[Iterable[str]] = None,
        barcodes: Optional[Iterable[str]] = None,
        timeout_seconds: Optional[float] = DEFAULT_PLATE_HTTP_TIMEOUT,
        returning: Optional[Iterable[str]] = None
    ) -> Optional[List[Plate]]:
        """
        Bulk get Plates.

        Bulk Get operations on large plates may take much longer than normal. The timeout_seconds
        parameter will use a higher HTTP timeout than the regular default. Pass a float to override
        it or pass None to use the standard client default.

        See https://benchling.com/api/reference#/Plates/bulkGetPlates
        """
        timeout_client = self.client.with_timeout(timeout_seconds) if timeout_seconds else self.client
        plate_id_string = optional_array_query_param(plate_ids)
        barcode_string = optional_array_query_param(barcodes)
        returning_string = optional_array_query_param(returning)
        response = bulk_get_plates.sync_detailed(
            client=timeout_client,
            plate_ids=none_as_unset(plate_id_string),
            barcodes=none_as_unset(barcode_string),
            returning=none_as_unset(returning_string),
        )
        plates_list = model_from_detailed(response)
        return plates_list.plates

    @api_method
    def create(self, plate: PlateCreate, returning: Optional[Iterable[str]] = None) -> Plate:
        """
        Create a Plate.

        See https://benchling.com/api/reference#/Plates/createPlate
        """
        returning_string = optional_array_query_param(returning)
        response = create_plate.sync_detailed(
            client=self.client, returning=none_as_unset(returning_string), json_body=plate
        )
        return model_from_detailed(response)

    @api_method
    def update(self, plate_id: str, plate: PlateUpdate, returning: Optional[Iterable[str]] = None) -> Plate:
        """
        Update a Plate.

        See https://benchling.com/api/reference#/Plates/updatePlate
        """
        returning_string = optional_array_query_param(returning)
        response = update_plate.sync_detailed(
            client=self.client, plate_id=plate_id, json_body=plate, returning=none_as_unset(returning_string)
        )
        return model_from_detailed(response)

    @api_method
    def archive(
        self, plate_ids: Iterable[str], reason: PlatesArchiveReason, should_remove_barcodes: bool
    ) -> PlatesArchivalChange:
        """
        Archive Plates.

        See https://benchling.com/api/reference#/Plates/archivePlates
        """
        archive_request = PlatesArchive(
            plate_ids=list(plate_ids), reason=reason, should_remove_barcodes=should_remove_barcodes
        )
        response = archive_plates.sync_detailed(client=self.client, json_body=archive_request)
        return model_from_detailed(response)

    @api_method
    def unarchive(self, plate_ids: Iterable[str]) -> PlatesArchivalChange:
        """
        Unarchive Plates.

        See https://benchling.com/api/reference#/Plates/unarchivePlates
        """
        unarchive_request = PlatesUnarchive(plate_ids=list(plate_ids))
        response = unarchive_plates.sync_detailed(client=self.client, json_body=unarchive_request)
        return model_from_detailed(response)



from time import sleep, time

from benchling_api_client.types import UNSET
from benchling_api_client.v2.alpha.extensions import (
    NotPresentError as NotPresentErrorAlpha,
    UnknownType as UnknownTypeAlpha,
)
from benchling_api_client.v2.beta.extensions import (
    NotPresentError as NotPresentErrorBeta,
    UnknownType as UnknownTypeBeta,
)
from benchling_api_client.v2.stable.api.tasks import get_task
from benchling_api_client.v2.stable.extensions import NotPresentError, UnknownType

from benchling_sdk.errors import WaitForTaskExpiredError
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.models import AsyncTask, AsyncTaskStatus
from benchling_sdk.services.v2.base_service import BaseService


class TaskService(BaseService):
    """
    Tasks.

    Endpoints that perform expensive computations launch long-running tasks. These endpoints return the task ID (a
    UUID) in the response body.

    After launching a task, periodically invoke the Get a task endpoint with the task UUID (e.g., every 10
    seconds), until the status is no longer RUNNING.

    You can access a task for up to 30 minutes after its completion, after which its data will no longer be
    available.

    See https://benchling.com/api/reference#/Tasks
    """

    @api_method
    def get_by_id(self, task_id: str) -> AsyncTask:
        """
        Get a Task by ID.

        See https://benchling.com/api/reference#/Tasks/getTask
        """
        response = get_task.sync_detailed(client=self.client, task_id=task_id)

        # TODO BNCH-50360 - AsyncTask is deserializing as UnknownType due to partial discriminator support
        # Undo this workaround once full discriminator support for tasks is available
        task_model = model_from_detailed(response)
        if isinstance(task_model, (UnknownTypeAlpha, UnknownTypeBeta, UnknownType)):
            return AsyncTask.from_dict(task_model.value)
        # Convert to a generic AsyncTask if a more specific type is returned
        return _deserialize_task_from_instance(task_model)

    @api_method
    def wait_for_task(
        self, task_id: str, interval_wait_seconds: int = 1, max_wait_seconds: int = 30
    ) -> AsyncTask:
        """
        Wait for a task until completion or timeout.

        A blocking method which polls the Benchling API and will return an AsyncTask as
        soon as its status is not RUNNING (in progress). This does not guarantee that the
        task was successful, only that Benchling has finished executing it.

        If max_wait_seconds is exceeded, will raise
        :py:class:`benchling_sdk.errors.WaitForTaskExpiredError`

        :param task_id: The ID of the task to poll
        :param interval_wait_seconds: time to wait between API calls in seconds
        :param max_wait_seconds: maximum wait time in seconds before raising WaitForTaskExpiredError
        :return: The completed AsyncTask. Check `status` for success or failure
        :rtype: AsyncTask
        """
        start_time = time()
        response = self.get_by_id(task_id)
        while response.status == AsyncTaskStatus.RUNNING and time() - start_time <= max_wait_seconds:
            sleep(interval_wait_seconds)
            response = self.get_by_id(task_id)
        if response.status != AsyncTaskStatus.RUNNING:
            return response
        raise WaitForTaskExpiredError(
            message=f"Timed out waiting for task ID {task_id} " f"after {max_wait_seconds} seconds",
            task=response,
        )


def _deserialize_task_from_instance(task_instance) -> AsyncTask:
    # TODO BNCH-50360 - AsyncTask is deserializing as UnknownType due to partial discriminator support
    # However, in some cases it will deserialize to a concrete type, often an incorrect type
    # Manually serialize to Async Task. Calling .from_dict() on its own results in an error:
    # AttributeError: 'NoneType' object has no attribute 'copy' from response being None

    # Successful task seem to deserialize to UnknownType but code defensively just in case
    try:
        if hasattr(task_instance, "response") and task_instance.response is not None:
            serialized = task_instance.to_dict()
            serialized["errors"] = UNSET
            return AsyncTask.from_dict(serialized)
    except (NotPresentErrorAlpha, NotPresentErrorBeta, NotPresentError):
        pass
    serialized = task_instance.to_dict()
    # AsyncErrors data is always in additional_properties,
    # It has no modeled properties of its own. They don't copy over properly so manually move them
    try:
        if hasattr(task_instance, "errors") and task_instance.errors is not None:
            serialized["errors"] = task_instance.errors.additional_properties
    except (NotPresentErrorAlpha, NotPresentErrorBeta, NotPresentError):
        pass
    serialized["response"] = UNSET
    return AsyncTask.from_dict(serialized)



from io import BytesIO
import mimetypes
from pathlib import Path
import sys
import tempfile
from typing import cast, Iterable, List, Optional, Union

from benchling_api_client.v2.stable.api.blobs import (
    abort_multipart_blob,
    bulk_get_blobs,
    complete_multipart_blob,
    create_blob,
    create_blob_part,
    create_multipart_blob,
    get_blob,
    get_blob_url,
)
import httpx
from typing_extensions import Protocol

from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.file_helpers import calculate_md5, encode_base64
from benchling_sdk.helpers.logging_helpers import sdk_logger
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import array_query_param
from benchling_sdk.models import (
    Blob,
    BlobComplete,
    BlobCreate,
    BlobCreateType,
    BlobMultipartCreate,
    BlobMultipartCreateType,
    BlobPart,
    BlobPartCreate,
    BlobUrl,
)
from benchling_sdk.services.v2.base_service import BaseService

DEFAULT_HTTP_TIMEOUT: float = 60.0
# 10 MB in bytes
MINIMUM_CHUNK_SIZE_BYTES: int = int(10e6)
DEFAULT_CHUNK_SIZE_BYTES: int = MINIMUM_CHUNK_SIZE_BYTES


def _write_blob_file(file_contents: BytesIO, file_path: Optional[Path] = None) -> Path:
    """Write bytes to a file. Will create a temp file if file_path isn't specified.

    Returns the Path of the file written to, whether created or used.
    """
    if file_path:
        with open(file_path, "wb") as existing_file_handle:
            existing_file_handle.write(file_contents.getbuffer())
        return file_path
    handle, temp_file_path = tempfile.mkstemp()
    with open(handle, "wb") as temp_file_handle:
        temp_file_handle.write(file_contents.getbuffer())
    return Path(temp_file_path)


class _BytesReader(Protocol):
    def __call__(self, chunk_size_bytes: int) -> bytes:
        pass


class BlobService(BaseService):
    """
    Blobs.

    Blobs are opaque files that can be linked to other items in Benchling, like assay runs or results. For example,
    you can upload a blob, then upload an assay result that links to that blob by ID. The blob will then appear as
    part of the assay result in the Benchling web UI.

    See https://benchling.com/api/reference#/Blobs
    """

    @api_method
    def get_by_id(self, blob_id: str) -> Blob:
        """
        Get a Blob.

        See https://benchling.com/api/reference#/Blobs/getBlob
        """
        response = get_blob.sync_detailed(client=self.client, blob_id=blob_id)
        return model_from_detailed(response)

    @api_method
    def create(self, blob: BlobCreate, timeout_seconds: float = DEFAULT_HTTP_TIMEOUT) -> Blob:
        """
        Upload a single-part blob.

        Blobs larger than 10MB should be uploaded in multiple parts. The data64 parameter is the
        base64-encoded part contents, and the md5 parameter is the hex-encoded MD5 hash of the part
        contents. For example, given the string hello, data64 is aGVsbG8= and md5 is 5d41402abc4b2a76b9719d911017c592

        See https://benchling.com/api/reference#/Blobs/createBlob
        """
        timeout_client = self.client.with_timeout(timeout_seconds)
        response = create_blob.sync_detailed(client=timeout_client, json_body=blob)
        return model_from_detailed(response)

    @api_method
    def create_from_bytes(
        self,
        input_bytes: Union[BytesIO, bytes],
        name: str,
        mime_type: Optional[str] = None,
        blob_type: BlobCreateType = BlobCreateType.RAW_FILE,
        timeout_seconds: float = DEFAULT_HTTP_TIMEOUT,
        chunk_size_bytes: int = DEFAULT_CHUNK_SIZE_BYTES,
    ) -> Blob:
        """Create a Benchling Blob from bytes or a BytesIO stream.

        Will automatically attempt a multi-part upload if the stream appears larger
        than the API's maximum size for single Blobs.

        :param input_bytes: The bytes or stream to upload
        :param name: The name of the Blob in Benchling
        :param mime_type: The representative MIME type for the Blob
        :param blob_type: The type of Blob in Benchling (Visualization or Raw File)
        :param timeout_seconds: Extends the normal HTTP timeout settings since Blob uploads can be large
            Use this to extend even further if streams are very large
        :param chunk_size_bytes: The size in bytes for each chunk when using a multipart upload. If the
            bytes exceed chunk_size_bytes, multipart will automatically be attempted. Otherwise, single
            Blob upload will be used
        :return: The created Blob
        :rtype: Blob
        """
        self.validate_chunk_size(chunk_size_bytes)
        if not isinstance(input_bytes, BytesIO):
            input_bytes = BytesIO(input_bytes)
        bytes_size = sys.getsizeof(input_bytes)
        chunk_producer: _BytesReader = lambda chunk_size: input_bytes.read(chunk_size)
        if bytes_size > MINIMUM_CHUNK_SIZE_BYTES:
            sdk_logger.info(
                "Size of input (%s bytes) exceeds minimum chunk size of %s bytes, "
                "initiating multipart upload",
                bytes_size,
                MINIMUM_CHUNK_SIZE_BYTES,
            )
            return self._multipart_from_bytes(
                chunk_producer=chunk_producer,
                name=name,
                blob_type=self._multipart_type(blob_type),
                mime_type=mime_type,
                timeout_seconds=timeout_seconds,
                chunk_size_bytes=chunk_size_bytes,
            )
        else:
            contents = input_bytes.read()
            return self._create_blob(
                blob_input=contents,
                name=name,
                blob_type=blob_type,
                mime_type=mime_type,
                timeout_seconds=timeout_seconds,
            )

    @api_method
    def create_from_file(
        self,
        file_path: Path,
        name: Optional[str] = None,
        mime_type: Optional[str] = None,
        blob_type: BlobCreateType = BlobCreateType.RAW_FILE,
        auto_detect: bool = True,
        timeout_seconds: float = DEFAULT_HTTP_TIMEOUT,
        chunk_size_bytes: int = DEFAULT_CHUNK_SIZE_BYTES,
    ) -> Blob:
        """Create a Benchling Blob from a file.

        Will automatically attempt a multi-part upload if the file appears larger
        than the API's maximum size for single Blobs.

        :param file_path: The Path to the file to upload
        :param name: The name of the Blob in Benchling
        :param mime_type: The representative mime type for the Blob
        :param blob_type: The type of Blob in Benchling (Visualization or Raw File)
        :param auto_detect: Will attempt to guess the file's MIME type if mime_type was not specified
            and auto_detect is True
        :param timeout_seconds: Extends the normal HTTP timeout settings since Blob uploads can be large
            Use this to extend even further if streams are very large
        :param chunk_size_bytes: The size in bytes for each chunk when using a multipart upload. If the
            bytes exceed chunk_size_bytes, multipart will automatically be attempted. Otherwise, single
            Blob upload will be used
        :return: The created Blob
        :rtype: Blob
        """
        self.validate_chunk_size(chunk_size_bytes)
        if not (file_path.exists() and file_path.is_file()):
            raise FileNotFoundError(f"{file_path} does not exist or is not a valid file")
        if auto_detect:
            if not name:
                name = file_path.name
            if not mime_type:
                mime_type = mimetypes.guess_type(str(file_path))[0]
        else:
            assert name, "name must be provided if auto_detect is not set to True"
        with open(file_path, "rb") as file:
            file_size_bytes = file_path.stat().st_size
            if file_size_bytes > MINIMUM_CHUNK_SIZE_BYTES:
                sdk_logger.info(
                    "Size of %s (%s bytes) exceeds minimum chunk size of %s bytes, "
                    "initiating multipart upload",
                    str(file_path),
                    file_size_bytes,
                    MINIMUM_CHUNK_SIZE_BYTES,
                )
                chunk_producer: _BytesReader = lambda chunk_size: file.read(chunk_size)
                return self._multipart_from_bytes(
                    chunk_producer=chunk_producer,
                    name=name,
                    blob_type=self._multipart_type(blob_type),
                    mime_type=mime_type,
                    timeout_seconds=timeout_seconds,
                    chunk_size_bytes=chunk_size_bytes,
                )
            else:
                file_contents = file.read()
                return self.create_from_bytes(
                    input_bytes=file_contents,
                    name=name,
                    blob_type=blob_type,
                    mime_type=mime_type,
                    timeout_seconds=timeout_seconds,
                )

    @api_method
    def download_url(self, blob_id: str) -> BlobUrl:
        """
        Get a Blob's download URL.

        See https://benchling.com/api/reference#/Blobs/getBlobUrl
        """
        response = get_blob_url.sync_detailed(client=self.client, blob_id=blob_id)
        return model_from_detailed(response)

    @api_method
    def download_bytes(self, blob_id: str) -> BytesIO:
        """Download a Blob in its entirety from Benchling and write it to BytesIO."""
        blob_download_wrapper = self.download_url(blob_id)
        download_url = cast(str, blob_download_wrapper.download_url)
        response = httpx.get(download_url)
        return BytesIO(response.content)

    @api_method
    def download_file(self, blob_id: str, destination_path: Optional[Path] = None) -> Path:
        """
        Download a Blob from Benchling and writes it to a file.

        If specified, destination_path must be a reference to a file and not a directory.
        It must also not already exist on the file system. If destination_path is None,
        a temporary file will be created for writing.
        """
        if destination_path and destination_path.is_dir():
            raise IsADirectoryError(
                f"Cannot write blob to directory '{destination_path}', specify a file instead"
            )
        elif destination_path and destination_path.exists():
            raise FileExistsError(f"Cannot write blob to '{destination_path}', the file already exists")
        blob_bytes = self.download_bytes(blob_id)
        destination_path = _write_blob_file(blob_bytes, destination_path)
        if not destination_path.exists():
            raise FileNotFoundError(f"Unable to write blob {blob_id} to '{destination_path}'")
        return destination_path

    @api_method
    def bulk_get(self, blob_ids: Iterable[str]) -> List[Blob]:
        """
        Bulk get Blobs by UUID.

        See https://benchling.com/api/reference#/Blobs/bulkGetBlobs
        """
        blob_ids_string = array_query_param(blob_ids)
        response = bulk_get_blobs.sync_detailed(client=self.client, blob_ids=blob_ids_string)
        results_list = model_from_detailed(response)
        return results_list.blobs

    @api_method
    def create_multipart_upload(self, multipart_blob: BlobMultipartCreate) -> Blob:
        """
        Initiate multi-part Blob upload.

        See https://benchling.com/api/reference#/Blobs/createMultipartBlob
        """
        response = create_multipart_blob.sync_detailed(client=self.client, json_body=multipart_blob)
        return model_from_detailed(response)

    @api_method
    def create_part(
        self, blob_id: str, blob_part: BlobPartCreate, timeout_seconds: float = DEFAULT_HTTP_TIMEOUT
    ) -> BlobPart:
        """
        Upload a part of a multi-part Blob.

        Larger files and slower connections will likely need to set a higher timeout_seconds value in
        order to complete successful part uploads.

        See https://benchling.com/api/reference#/Blobs/createBlobPart
        """
        timeout_client = self.client.with_timeout(timeout_seconds)
        response = create_blob_part.sync_detailed(client=timeout_client, blob_id=blob_id, json_body=blob_part)
        return model_from_detailed(response)

    @api_method
    def complete_multipart_upload(self, blob_id: str, blob_parts: Iterable[BlobPart]) -> Blob:
        """
        Combine Blob parts into a single Blob.

        See https://benchling.com/api/reference#/Blobs/completeMultipartBlob
        """
        blob_complete = BlobComplete(parts=list(blob_parts))
        response = complete_multipart_blob.sync_detailed(
            client=self.client, blob_id=blob_id, json_body=blob_complete
        )
        return model_from_detailed(response)

    @api_method
    def abort_multipart_upload(self, blob_id: str) -> None:
        """
        Abort multi-part blob upload.

        See https://benchling.com/api/reference#/Blobs/abortMultipartBlob
        """
        response = abort_multipart_blob.sync_detailed(client=self.client, blob_id=blob_id)
        # Even though we won't return, will do the other processing such as status checking
        empty_object = model_from_detailed(response)  # noqa: F841

    def _create_blob(
        self,
        blob_input: bytes,
        name: str,
        blob_type: BlobCreateType,
        mime_type: Optional[str],
        timeout_seconds: float,
    ) -> Blob:
        data = encode_base64(blob_input)
        md5 = calculate_md5(blob_input)
        blob_create = BlobCreate(name=name, type=blob_type, data64=data, md5=md5)
        # Use BlobCreate's default mime type unless we specified one
        if mime_type is not None:
            blob_create.mime_type = mime_type
        return self.create(blob=blob_create, timeout_seconds=timeout_seconds)

    def _multipart_from_bytes(
        self,
        chunk_producer: _BytesReader,
        name: str,
        mime_type: Optional[str],
        blob_type: BlobMultipartCreateType,
        timeout_seconds: float,
        chunk_size_bytes: int,
    ) -> Blob:
        blob_parts = []
        part_number = 0
        multipart_args = {"name": name, "type": blob_type}
        # Do not override BlobMultipartCreate's mime_type unless explicitly set
        if mime_type:
            multipart_args["mime_type"] = mime_type
        multipart_blob = BlobMultipartCreate(**multipart_args)  # type: ignore
        start_blob = self.create_multipart_upload(multipart_blob=multipart_blob)
        try:
            while True:
                cursor = chunk_producer(chunk_size_bytes)
                if not cursor:
                    break
                part_number += 1
                sdk_logger.debug("Beginning upload of part %s for blob '%s'", part_number, name)
                created_part = self._upload_part(
                    parent_blob_id=start_blob.id,
                    part_number=part_number,
                    input_bytes=cursor,
                    timeout_seconds=timeout_seconds,
                )
                blob_parts.append(created_part)
            sdk_logger.info("Uploaded %s total parts of blob '%s', completing upload", part_number, name)
            completed_blob = self.complete_multipart_upload(blob_id=start_blob.id, blob_parts=blob_parts)
            return completed_blob
        except Exception as e:
            sdk_logger.error(
                "Error while uploading part %s for blob '%s', aborting upload", part_number, name
            )
            self.abort_multipart_upload(blob_id=start_blob.id)
            raise e

    def _upload_part(
        self, parent_blob_id: str, part_number: int, input_bytes: bytes, timeout_seconds: float
    ) -> BlobPart:
        data = encode_base64(input_bytes)
        md5 = calculate_md5(input_bytes)
        blob_part = BlobPartCreate(part_number=part_number, data64=data, md5=md5)
        created_part = self.create_part(
            blob_id=parent_blob_id, blob_part=blob_part, timeout_seconds=timeout_seconds
        )
        return created_part

    @staticmethod
    def validate_chunk_size(chunk_size_bytes: int):
        """Validate the specified bytes chunk size."""
        if chunk_size_bytes < MINIMUM_CHUNK_SIZE_BYTES:
            raise ValueError(
                f"chunk_size_bytes was {chunk_size_bytes}. "
                f"When specified, it must be at least the "
                f"minimum chunk size of {MINIMUM_CHUNK_SIZE_BYTES}"
            )

    @staticmethod
    def _multipart_type(blob_type: BlobCreateType):
        # TODO These models should probably be the same
        return BlobMultipartCreateType(blob_type.name)



from benchling_api_client.v2.alpha.api.dna_sequences import find_matching_regions_dna_sequences
from benchling_api_client.v2.alpha.api.dna_sequences.find_matching_regions_dna_sequences import (
    DnaSequencesFindMatchingRegion,
)

from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.models import AsyncTaskLink
from benchling_sdk.services.v2.base_service import BaseService


class V2AlphaDnaSequenceService(BaseService):
    """
    V2-Alpha DNA Sequences.

    DNA sequences are the bread and butter of the Benchling Molecular Biology suite. On Benchling, these are
    comprised of a string of nucleotides and collections of other attributes, such as annotations and primers.

    See https://benchling.com/api/v2-alpha/reference#/DNA%20Sequences
    """

    @api_method
    def find_matching_regions(self, find_matching_region: DnaSequencesFindMatchingRegion) -> AsyncTaskLink:
        """
        Find matching regions for DNA sequences.

        See https://benchling.com/api/v2-alpha/reference#/DNA%20Sequences/findMatchingRegionsDnaSequences
        """
        response = find_matching_regions_dna_sequences.sync_detailed(
            client=self.client, json_body=find_matching_region
        )
        return model_from_detailed(response)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, Optional, Type, TypeVar, Union

import attr
from benchling_api_client.v2.benchling_client import AuthorizationMethod
from benchling_api_client.webhooks.v0.alpha.models.webhook_envelope import (
    WebhookEnvelope as WebhookEnvelopeAlpha,
)
from benchling_api_client.webhooks.v0.beta.models.webhook_envelope import (
    WebhookEnvelope as WebhookEnvelopeBeta,
)
import httpx
from typing_extensions import Protocol

from benchling_sdk.apps import helpers
from benchling_sdk.apps.config.decryption_provider import BaseDecryptionProvider
from benchling_sdk.apps.config.dependencies import BaseDependencies
from benchling_sdk.benchling import (
    _DEFAULT_BASE_PATH,
    _DEFAULT_RETRY_STRATEGY,
    Benchling,
    BenchlingApiClientDecorator,
)
from benchling_sdk.helpers.logging_helpers import log_stability_warning, StabilityLevel
from benchling_sdk.helpers.retry_helpers import RetryStrategy

log_stability_warning(StabilityLevel.ALPHA)


ConfigType = TypeVar("ConfigType", bound=BaseDependencies)
AppType = TypeVar("AppType", bound="App")
AppWebhookType = Union[WebhookEnvelopeAlpha, WebhookEnvelopeBeta]


class MissingTenantUrlProviderError(Exception):
    """Error when a base URL is expected but unspecified."""

    pass


class MissingAppConfigTypeError(Exception):
    """Error when app config is expected but unspecified."""

    pass


class MalformedAppWebhookError(Exception):
    """Error when a webhook cannot be read by an app."""

    pass


class TenantUrlProvider(Protocol):
    """Return a base URL."""

    def __call__(self) -> str:
        """Return a base URL."""
        pass


def tenant_url_provider_static(tenant_url: str) -> TenantUrlProvider:
    """Create a provider function that always returns a static tenant URL."""

    def _url() -> str:
        return tenant_url

    return _url


def tenant_url_provider_lazy() -> TenantUrlProvider:
    """
    Create a provider function for app that will be initialized at runtime, such as from a webhook.

    Useful for when a base_url for Benchling is not known in advance but can be supplied at runtime.
    """

    def _deferred() -> str:
        raise MissingTenantUrlProviderError(
            "Unable to initialize base URL for tenant. Expected a URL to "
            "be provided at runtime but none was specified. Either specify "
            "a url provider or use TenantUrlProvider.static_url"
        )

    return _deferred


class BenchlingProvider(Protocol):
    """Return a Benchling instance."""

    def __call__(self, tenant_url_provider: TenantUrlProvider) -> Benchling:
        """Return a Benchling instance."""
        pass


class ConfigProvider(Protocol[ConfigType]):
    """Return a ConfigType instance."""

    def __call__(self, app: App[ConfigType]) -> ConfigType:
        """Return a ConfigType instance."""
        pass


def config_provider_static(config: ConfigType) -> ConfigProvider[ConfigType]:
    """Create a provider function that always returns a static app config."""

    def _static_config(app: App[ConfigType]) -> ConfigType:
        return config

    return _static_config


def config_provider_error_on_call() -> ConfigProvider[ConfigType]:
    """
    Create a provider function that raises an error.

    Used as a ConfigProvider for apps which don't support config and don't expect to invoke it.
    """

    def _error_on_call(app: App[ConfigType]) -> ConfigType:
        raise MissingAppConfigTypeError(
            "No app config class was defined for this app. "
            "Initialize an app with a ConfigProvider to use config."
        )

    return _error_on_call


def benchling_provider_static(benchling: Benchling) -> BenchlingProvider:
    """Create a provider function that always returns a static Benchling."""

    def _static_benchling(tenant_url_provider: TenantUrlProvider) -> Benchling:
        return benchling

    return _static_benchling


@attr.s(auto_attribs=True)
class App(Generic[ConfigType]):
    """
    App.

    See https://docs.benchling.com/docs/getting-started-benchling-apps

    Accepts providers as arguments to lazily initialize since some required attributes may not be
    known until runtime. Also allows for easier mocking in tests.
    """

    id: str
    _benchling_provider: BenchlingProvider
    _tenant_url_provider: TenantUrlProvider
    _config_provider: ConfigProvider[ConfigType] = attr.ib(default=config_provider_error_on_call)
    _benchling: Optional[Benchling] = attr.ib(default=None, init=False)
    _config: Optional[ConfigType] = attr.ib(default=None, init=False)

    @property
    def benchling(self) -> Benchling:
        """Return a Benchling instance for the App."""
        if self._benchling is None:
            self._benchling = self._benchling_provider(self._tenant_url_provider)
        return self._benchling

    @property
    def config(self) -> ConfigType:
        """
        Return config for the app.

        Apps which do not have config will raise MissingAppConfigTypeError.
        """
        if self._config is None:
            self._config = self._config_provider(self)
        return self._config

    def reset(self) -> None:
        """
        Reset the app.

        Generally clears all states and internal caches, which may cause subsequent invocations of the App
        to be expensive.
        """
        self._benchling = None
        if self._config is not None:
            self._config.invalidate_cache()

    def with_base_url(self: AppType, base_url: str) -> AppType:
        """Create a new copy of the app with a different base URL."""
        updated_tenant_url_provider = tenant_url_provider_static(base_url)
        modified_app = attr.evolve(self, tenant_url_provider=updated_tenant_url_provider)
        modified_app.reset()
        return modified_app

    def with_webhook(self: AppType, webhook: Union[dict, AppWebhookType]) -> AppType:
        """Create a new copy of the app with a different base URL provided by a webhook."""
        if isinstance(webhook, dict):
            if "baseUrl" not in webhook:
                raise MalformedAppWebhookError("The webhook specified did not contain a baseUrl")
            base_url = webhook["baseUrl"]
        else:
            base_url = webhook.base_url
        return self.with_base_url(base_url)

    def create_session_context(
        self: AppType,
        name: str,
        timeout_seconds: int,
        context_enter_handler: Optional[helpers.session_helpers.SessionContextEnterHandler[AppType]] = None,
        context_exit_handler: Optional[helpers.session_helpers.SessionContextExitHandler[AppType]] = None,
    ) -> helpers.session_helpers.SessionContextManager[AppType]:
        """
        Create Session Context.

        Create a new app session in Benchling.
        """
        # Avoid circular import + MyPy "is not defined" if using relative like above
        from benchling_sdk.apps.helpers.session_helpers import new_session_context

        return new_session_context(self, name, timeout_seconds, context_enter_handler, context_exit_handler)

    def continue_session_context(
        self: AppType,
        session_id: str,
        context_enter_handler: Optional[helpers.session_helpers.SessionContextEnterHandler[AppType]] = None,
        context_exit_handler: Optional[helpers.session_helpers.SessionContextExitHandler[AppType]] = None,
    ) -> helpers.session_helpers.SessionContextManager[AppType]:
        """
        Continue Session Context.

        Fetch an existing app session from Benchling and enter a context with it.
        """
        # Avoid circular import + MyPy "is not defined" if using relative like above
        from benchling_sdk.apps.helpers.session_helpers import continue_session_context

        return continue_session_context(self, session_id, context_enter_handler, context_exit_handler)

    @classmethod
    def init(
        cls: Type[AppType],
        id: str,
        benchling_provider: BenchlingProvider,
        tenant_url_provider: TenantUrlProvider,
        config_provider: Optional[ConfigProvider] = None,
    ) -> AppType:
        """
        Init.

        Initialize an app from its class.
        """
        required_config_provider: ConfigProvider[ConfigType] = (
            config_provider_error_on_call() if config_provider is None else config_provider
        )
        return cls(id, benchling_provider, tenant_url_provider, required_config_provider)


class BaseAppFactory(ABC, Generic[AppType, ConfigType]):
    """
    Base App Factory.

    Can be used as an alternative to init_app() for those who prefer to import a pre-defined app instance
    globally. Call create() on the factory to initialize an App.

    Users must subclass AppFactory and implement its abstract methods to create a subclass of App.
    """

    _app_type: Type[AppType]
    app_id: str
    benchling_provider: BenchlingProvider
    config_provider: ConfigProvider[ConfigType]

    def __init__(self, app_type: Type[AppType], app_id: str, config_type: Optional[Type[ConfigType]] = None):
        """Initialize App Factory."""
        self._app_type = app_type
        self.app_id = app_id

        # Initialize providers here to fail fast if there is a problem assembling them from the factory

        def benchling_provider(tenant_url_provider):
            tenant_url = tenant_url_provider()
            return Benchling(
                url=tenant_url,
                auth_method=self.auth_method,
                base_path=self.base_path,
                retry_strategy=self.retry_strategy,
                client_decorator=self.client_decorator,
                httpx_client=self.httpx_client,
            )

        if config_type is None:
            config_provider: ConfigProvider[ConfigType] = config_provider_error_on_call()
        else:

            def _config_provider(app: App[ConfigType]) -> ConfigType:
                # MyPy believes config_type can be None despite the conditional
                return config_type.from_app(app.benchling, app.id, self.decryption_provider)  # type: ignore

            config_provider = _config_provider

        self.benchling_provider = benchling_provider
        self.config_provider = config_provider

    def create(self) -> AppType:
        """Create an App instance from the factory."""
        return self._app_type.init(
            self.app_id, self.benchling_provider, self.tenant_url_provider, self.config_provider
        )

    @property
    @abstractmethod
    def auth_method(self) -> AuthorizationMethod:
        """
        Get an auth method to pass to Benchling.

        Must be implemented on all subclasses.
        """
        pass

    @property
    def tenant_url_provider(self) -> TenantUrlProvider:
        """
        Get a tenant URL provider that will provide a base URL for Benchling at runtime.

        By default, assumes that the App has no base_url and will be provided one later (e.g., from a webhook).
        Invoking app.benchling on an App in this state without setting a URL will raise an error.

        Use tenant_url_provider_static("https://myurl...") to specify a single URL.
        """
        return tenant_url_provider_lazy()

    # Benchling overrides

    @property
    def base_path(self) -> Optional[str]:
        """Get a base_path for Benchling."""
        return _DEFAULT_BASE_PATH

    @property
    def client_decorator(self) -> Optional[BenchlingApiClientDecorator]:
        """Get a BenchlingApiClientDecorator for Benchling."""
        return None

    @property
    def httpx_client(self) -> Optional[httpx.Client]:
        """Get a custom httpx Client for Benchling."""
        return None

    @property
    def retry_strategy(self) -> RetryStrategy:
        """Get a RetryStrategy for Benchling."""
        return _DEFAULT_RETRY_STRATEGY

    @property
    def decryption_provider(self) -> Optional[BaseDecryptionProvider]:
        """Get a decryption provider for decryption app config secrets."""
        return None


def init_app(
    app_id: str,
    benchling_provider: BenchlingProvider,
    tenant_url_provider: TenantUrlProvider,
    config_provider: Optional[ConfigProvider[ConfigType]] = None,
) -> App[ConfigType]:
    """
    Init App.

    Initializes a Benchling App with a series of functions to provide App dependencies at runtime.
    """
    if config_provider is None:
        config_provider = config_provider_error_on_call()
    return App(app_id, benchling_provider, tenant_url_provider, config_provider)


def init_static_app(
    app_id: str, benchling: Benchling, config: Optional[ConfigType] = None
) -> App[ConfigType]:
    """
    Init Static App.

    Initializes a Benchling App with static values. Suitable for apps that communicate with a single URL.
    """
    tenant_url_provider = tenant_url_provider_static(benchling.client.base_url)
    benchling_provider = benchling_provider_static(benchling)
    config_provider = config_provider_error_on_call() if config is None else config_provider_static(config)
    return init_app(app_id, benchling_provider, tenant_url_provider, config_provider)



from datetime import date, datetime
from typing import List, Optional, Type, Union

from benchling_api_client.v2.beta.models.app_config_field_type import AppConfigFieldType
from benchling_api_client.v2.beta.models.base_manifest_config import BaseManifestConfig
from benchling_api_client.v2.beta.models.dropdown_dependency import DropdownDependency
from benchling_api_client.v2.beta.models.dropdown_dependency_types import DropdownDependencyTypes
from benchling_api_client.v2.beta.models.entity_schema_dependency import EntitySchemaDependency
from benchling_api_client.v2.beta.models.entity_schema_dependency_type import EntitySchemaDependencyType
from benchling_api_client.v2.beta.models.field_definitions_manifest import FieldDefinitionsManifest
from benchling_api_client.v2.beta.models.manifest_array_config import ManifestArrayConfig
from benchling_api_client.v2.beta.models.manifest_scalar_config import ManifestScalarConfig
from benchling_api_client.v2.beta.models.resource_dependency import ResourceDependency
from benchling_api_client.v2.beta.models.resource_dependency_types import ResourceDependencyTypes
from benchling_api_client.v2.beta.models.scalar_config import ScalarConfig
from benchling_api_client.v2.beta.models.scalar_config_types import ScalarConfigTypes
from benchling_api_client.v2.beta.models.schema_dependency import SchemaDependency
from benchling_api_client.v2.beta.models.schema_dependency_subtypes import SchemaDependencySubtypes
from benchling_api_client.v2.beta.models.schema_dependency_types import SchemaDependencyTypes
from benchling_api_client.v2.beta.models.workflow_task_schema_dependency import WorkflowTaskSchemaDependency
from benchling_api_client.v2.beta.models.workflow_task_schema_dependency_output import (
    WorkflowTaskSchemaDependencyOutput,
)
from benchling_api_client.v2.beta.models.workflow_task_schema_dependency_type import (
    WorkflowTaskSchemaDependencyType,
)
from benchling_api_client.v2.extensions import UnknownType
from benchling_api_client.v2.stable.extensions import NotPresentError

from benchling_sdk.apps.config.dependencies import ConfigurationReference, UnsupportedDependencyError
from benchling_sdk.apps.config.scalars import JsonType, ScalarModelType
from benchling_sdk.models import (
    AaSequence,
    ArrayElementAppConfigItem,
    AssayResult,
    AssayRun,
    BooleanAppConfigItem,
    Box,
    Container,
    CustomEntity,
    DateAppConfigItem,
    DatetimeAppConfigItem,
    DnaOligo,
    DnaSequence,
    EntitySchemaAppConfigItem,
    Entry,
    FieldAppConfigItem,
    FloatAppConfigItem,
    GenericApiIdentifiedAppConfigItem,
    IntegerAppConfigItem,
    JsonAppConfigItem,
    Location,
    Mixture,
    Molecule,
    Plate,
    Request,
    RnaOligo,
    RnaSequence,
    SecureTextAppConfigItem,
    TextAppConfigItem,
    WorkflowTask,
)

_MODEL_TYPES_FROM_SCHEMA_TYPE = {
    SchemaDependencyTypes.CONTAINER_SCHEMA: Container,
    SchemaDependencyTypes.PLATE_SCHEMA: Plate,
    SchemaDependencyTypes.BOX_SCHEMA: Box,
    SchemaDependencyTypes.LOCATION_SCHEMA: Location,
    SchemaDependencyTypes.ENTRY_SCHEMA: Entry,
    SchemaDependencyTypes.REQUEST_SCHEMA: Request,
    SchemaDependencyTypes.RESULT_SCHEMA: AssayResult,
    SchemaDependencyTypes.RUN_SCHEMA: AssayRun,
    SchemaDependencyTypes.WORKFLOW_TASK_SCHEMA: WorkflowTask,
}


_SCALAR_TYPES_FROM_CONFIG = {
    ScalarConfigTypes.BOOLEAN: bool,
    ScalarConfigTypes.DATE: date,
    ScalarConfigTypes.DATETIME: datetime,
    ScalarConfigTypes.FLOAT: float,
    ScalarConfigTypes.INTEGER: int,
    ScalarConfigTypes.JSON: JsonType,
    ScalarConfigTypes.TEXT: str,
}


_FIELD_SCALAR_TYPES_FROM_CONFIG = {
    AppConfigFieldType.BOOLEAN: bool,
    AppConfigFieldType.DATE: date,
    AppConfigFieldType.DATETIME: datetime,
    AppConfigFieldType.FLOAT: float,
    AppConfigFieldType.INTEGER: int,
    AppConfigFieldType.JSON: JsonType,
    AppConfigFieldType.TEXT: str,
}


ModelType = Union[AssayResult, AssayRun, Box, Container, Entry, Location, Plate, Request]

_INSTANCE_FROM_SCHEMA_SUBTYPE = {
    SchemaDependencySubtypes.AA_SEQUENCE: AaSequence,
    SchemaDependencySubtypes.CUSTOM_ENTITY: CustomEntity,
    SchemaDependencySubtypes.DNA_SEQUENCE: DnaSequence,
    SchemaDependencySubtypes.DNA_OLIGO: DnaOligo,
    SchemaDependencySubtypes.MIXTURE: Mixture,
    SchemaDependencySubtypes.MOLECULE: Molecule,
    SchemaDependencySubtypes.RNA_OLIGO: RnaOligo,
    SchemaDependencySubtypes.RNA_SEQUENCE: RnaSequence,
}

EntitySubtype = Union[
    AaSequence, CustomEntity, DnaOligo, DnaSequence, Mixture, Molecule, RnaOligo, RnaSequence
]

AnyDependency = Union[
    BaseManifestConfig,
    DropdownDependency,
    EntitySchemaDependency,
    FieldDefinitionsManifest,
    ManifestArrayConfig,
    ManifestScalarConfig,
    ResourceDependency,
    SchemaDependency,
    WorkflowTaskSchemaDependency,
]

ArrayElementDependency = Union[
    SchemaDependency,
    EntitySchemaDependency,
    WorkflowTaskSchemaDependency,
    DropdownDependency,
    ResourceDependency,
    ManifestScalarConfig,
]


class UnsupportedSubTypeError(Exception):
    """Error when an unsupported subtype is encountered."""

    pass


def model_type_from_dependency(
    dependency: Union[EntitySchemaDependency, SchemaDependency]
) -> Optional[Type[Union[ModelType, EntitySubtype]]]:
    """Translate a schema dependency to its model class. Must have a valid subtype."""
    if isinstance(dependency, EntitySchemaDependency):
        subtype = subtype_from_entity_schema_dependency(dependency)
        if subtype:
            return _subtype_instance_from_dependency(dependency)
        return None
    return _MODEL_TYPES_FROM_SCHEMA_TYPE[dependency.type]


def scalar_type_from_config(config: ScalarConfig) -> Union[object, Type[ScalarModelType]]:
    """Translate a scalar config to its Pyton type."""
    # We union with object to satisfy JsonType.
    return _SCALAR_TYPES_FROM_CONFIG.get(config.type, str)


def scalar_type_from_field_config(config: FieldDefinitionsManifest) -> Union[object, Type[ScalarModelType]]:
    """Translate a field config to its Pyton type."""
    # type may not be specified, so handle NotPresentError
    try:
        if hasattr(config, "type"):
            # We union with object to satisfy JsonType.
            return _FIELD_SCALAR_TYPES_FROM_CONFIG.get(config.type, str)
    # We can't seem to handle this programmatically by checking isinstance() or output truthiness
    except NotPresentError:
        pass
    return str


def field_definitions_from_dependency(
    dependency: Union[
        EntitySchemaDependency,
        SchemaDependency,
        WorkflowTaskSchemaDependency,
        WorkflowTaskSchemaDependencyOutput,
    ]
) -> List[FieldDefinitionsManifest]:
    """Safely return a list of field definitions from a schema dependency or empty list."""
    try:
        if hasattr(dependency, "field_definitions"):
            return dependency.field_definitions
    # We can't seem to handle this programmatically by checking isinstance() or field truthiness
    except NotPresentError:
        pass
    return []


def element_definition_from_dependency(dependency: ManifestArrayConfig) -> List[ArrayElementDependency]:
    """Safely return an element definition as a list of dependencies from an array dependency or empty list."""
    try:
        if hasattr(dependency, "element_definition"):
            return [
                _fix_element_definition_deserialization(element) for element in dependency.element_definition
            ]
    # We can't seem to handle this programmatically by checking isinstance() or field truthiness
    except NotPresentError:
        pass
    return []


# TODO BNCH-57036 All element definitions currently deserialize to UnknownType. Hack around this temporarily
def _fix_element_definition_deserialization(
    element: Union[UnknownType, ArrayElementDependency]
) -> ArrayElementDependency:
    if isinstance(element, UnknownType):
        if "type" in element.value:
            element_type = element.value["type"]
            if element_type == WorkflowTaskSchemaDependencyType.WORKFLOW_TASK_SCHEMA:
                return WorkflowTaskSchemaDependency.from_dict(element.value)
            elif element_type == EntitySchemaDependencyType.ENTITY_SCHEMA:
                return EntitySchemaDependency.from_dict(element.value)
            elif element_type in [member.value for member in SchemaDependencyTypes]:
                return SchemaDependency.from_dict(element.value)
            elif element_type == DropdownDependencyTypes.DROPDOWN:
                return DropdownDependency.from_dict(element.value)
            elif element_type in [member.value for member in ResourceDependencyTypes]:
                return ResourceDependency.from_dict(element.value)
            elif element_type in [member.value for member in ScalarConfigTypes]:
                return ManifestScalarConfig.from_dict(element.value)
        raise NotImplementedError(f"No array deserialization fix for {element}")
    return element


def workflow_task_schema_output_from_dependency(
    dependency: WorkflowTaskSchemaDependency,
) -> Optional[WorkflowTaskSchemaDependencyOutput]:
    """Safely return a workflow task schema output from a workflow task schema or None."""
    try:
        if hasattr(dependency, "output"):
            return dependency.output
    # We can't seem to handle this programmatically by checking isinstance() or output truthiness
    except NotPresentError:
        pass
    return None


def options_from_dependency(dependency: DropdownDependency) -> List[BaseManifestConfig]:
    """Safely return a list of options from a dropdown dependency or empty list."""
    try:
        if hasattr(dependency, "options"):
            return dependency.options
    # We can't seem to handle this programmatically by checking isinstance() or field truthiness
    except NotPresentError:
        pass
    return []


def is_config_required(dependency: AnyDependency) -> bool:
    """Safely return if a config item is required."""
    try:
        if hasattr(dependency, "required_config"):
            # `type: ignore` is used here because casting would be inaccurate or fragile.
            return dependency.required_config  # type: ignore
    # We can't seem to handle this programmatically by checking isinstance() or field truthiness
    except NotPresentError:
        pass
    return False


def is_field_value_required(dependency: FieldDefinitionsManifest) -> bool:
    """
    Safely return if a field must have a value.

    A field must be specified as requiredConfig: true AND isRequired: true to require a value.
    """
    # Fields must be both linked and isRequired to always provide a value
    try:
        if hasattr(dependency, "required_config") and hasattr(dependency, "is_required"):
            # dependency.required_config is Optional so evaluate the bool value
            return dependency.required_config is True and dependency.is_required is True
    # We can't seem to handle this programmatically by checking isinstance() or field truthiness
    except NotPresentError:
        pass
    return False


# For forwards compatibility with multi-valued scalars, which could infer Unset == False
def is_config_multi_valued(dependency: FieldDefinitionsManifest) -> bool:
    """
    Safely return if a config item is multi-valued.

    Assumes False in the case that a multi-valued constraint is unspecified (Unset).
    """
    multi_valued_unset = is_config_multi_valued_or_unset(dependency)
    return multi_valued_unset if multi_valued_unset is not None else False


def is_config_multi_valued_or_unset(dependency: FieldDefinitionsManifest) -> Optional[bool]:
    """
    Safely return if a config item is multi-valued or None if it's Unset.

    Multi-valued constraint being undefined cannot be safely inferred for type coercion.

    For instance, a required text field could be typed as:
    isMulti == True: List[str]
    isMulti == False: str
    isMulti == Unset: Union[str, List[str]]
    """
    try:
        if hasattr(dependency, "is_multi") and dependency.is_multi is not None:
            return dependency.is_multi
    # We can't seem to handle this programmatically by checking isinstance() or field truthiness
    except NotPresentError:
        pass
    return None


def subtype_from_entity_schema_dependency(
    dependency: EntitySchemaDependency,
) -> Optional[SchemaDependencySubtypes]:
    """Safely return an entity schema dependency's subtype, if present."""
    try:
        if hasattr(dependency, "subtype") and dependency.subtype:
            return dependency.subtype
    # We can't seem to handle this programmatically by checking isinstance() or field truthiness
    except NotPresentError:
        pass
    return None


def app_config_type_from_dependency(dependency: AnyDependency) -> Type[ConfigurationReference]:
    """
    App Config Item Type From Item.

    Returns the type of the API model corresponding to an app config item.
    Raises UnsupportedDependencyError if encountering an unknown type.
    """
    # Generic type
    if isinstance(
        dependency,
        (
            BaseManifestConfig,
            DropdownDependency,
            SchemaDependency,
            ResourceDependency,
            WorkflowTaskSchemaDependency,
        ),
    ):
        return GenericApiIdentifiedAppConfigItem
    # Specially handled Schema types
    elif isinstance(dependency, FieldDefinitionsManifest):
        return FieldAppConfigItem
    elif isinstance(dependency, EntitySchemaDependency):
        return EntitySchemaAppConfigItem
    # Scalar types
    elif isinstance(dependency, ManifestScalarConfig):
        if dependency.type == ScalarConfigTypes.BOOLEAN:
            return BooleanAppConfigItem
        elif dependency.type == ScalarConfigTypes.DATE:
            return DateAppConfigItem
        elif dependency.type == ScalarConfigTypes.DATETIME:
            return DatetimeAppConfigItem
        elif dependency.type == ScalarConfigTypes.FLOAT:
            return FloatAppConfigItem
        elif dependency.type == ScalarConfigTypes.INTEGER:
            return IntegerAppConfigItem
        elif dependency.type == ScalarConfigTypes.JSON:
            return JsonAppConfigItem
        elif dependency.type == ScalarConfigTypes.SECURE_TEXT:
            return SecureTextAppConfigItem
        elif dependency.type == ScalarConfigTypes.TEXT:
            return TextAppConfigItem
    elif isinstance(dependency, ManifestArrayConfig):
        return ArrayElementAppConfigItem
    raise UnsupportedDependencyError(f"Unrecognized type for {dependency}]")


def _subtype_instance_from_dependency(dependency: EntitySchemaDependency) -> Type[EntitySubtype]:
    if dependency.subtype in _INSTANCE_FROM_SCHEMA_SUBTYPE:
        return _INSTANCE_FROM_SCHEMA_SUBTYPE[dependency.subtype]
    # This would mean the spec has a new subtype, the manifest installed in Benchling has declared it,
    # the user has linked it in Benchling, but the app code receiving it was never updated.
    # App developers should support it in deployed app code before republishing the manifest.
    raise UnsupportedSubTypeError(
        f"An unsupported subtype, {dependency.subtype.value} was received. "
        f"The version of the SDK may need to be upgraded to support this."
    )



from typing import Any, Dict, Iterable, List, Optional, Union

from benchling_api_client.v2.stable.api.rna_sequences import (
    archive_rna_sequences,
    auto_annotate_rna_sequences,
    autofill_rna_sequence_parts,
    autofill_rna_sequence_translations,
    bulk_create_rna_sequences,
    bulk_get_rna_sequences,
    bulk_update_rna_sequences,
    create_rna_sequence,
    get_rna_sequence,
    list_rna_sequences,
    unarchive_rna_sequences,
    update_rna_sequence,
)
from benchling_api_client.v2.stable.models.autofill_rna_sequences import AutofillRnaSequences
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import (
    none_as_unset,
    optional_array_query_param,
    schema_fields_query_param,
)
from benchling_sdk.models import (
    AsyncTaskLink,
    AutoAnnotateRnaSequences,
    EntityArchiveReason,
    ListRNASequencesSort,
    RnaSequence,
    RnaSequenceBulkCreate,
    RnaSequenceBulkUpdate,
    RnaSequenceCreate,
    RnaSequencesArchivalChange,
    RnaSequencesArchive,
    RnaSequencesBulkCreateRequest,
    RnaSequencesBulkUpdateRequest,
    RnaSequencesPaginatedList,
    RnaSequencesUnarchive,
    RnaSequenceUpdate,
)
from benchling_sdk.services.v2.base_service import BaseService


class RnaSequenceService(BaseService):
    """
    RNA Sequences.

    Chains of linear, single stranded RNA that support most capabilities and attributes of DNA Sequences.

    See https://benchling.com/api/reference?stability=not-available#/RNA%20Sequences
    """

    @api_method
    def get_by_id(self, rna_sequence_id: str) -> RnaSequence:
        """
        Get a RNA sequence.

        See https://benchling.com/api/reference#/RNA%20Sequences/getRNASequence
        """
        response = get_rna_sequence.sync_detailed(client=self.client, rna_sequence_id=rna_sequence_id)
        return model_from_detailed(response)

    @api_method
    def _rna_sequences_page(
        self,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        bases: Optional[str] = None,
        folder_id: Optional[str] = None,
        mentioned_in: Optional[List[str]] = None,
        project_id: Optional[str] = None,
        registry_id: Optional[str] = None,
        schema_id: Optional[str] = None,
        archive_reason: Optional[str] = None,
        mentions: Optional[List[str]] = None,
        sort: Optional[ListRNASequencesSort] = None,
        ids: Optional[Iterable[str]] = None,
        entity_registry_ids_any_of: Optional[Iterable[str]] = None,
        name_includes: Optional[str] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        creator_ids: Optional[Iterable[str]] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
        page_size: Optional[int] = None,
        next_token: NextToken = None,
        author_idsany_of: Optional[Iterable[str]] = None,
    ) -> Response[RnaSequencesPaginatedList]:
        response = list_rna_sequences.sync_detailed(
            client=self.client,
            modified_at=none_as_unset(modified_at),
            name=none_as_unset(name),
            bases=none_as_unset(bases),
            folder_id=none_as_unset(folder_id),
            mentioned_in=none_as_unset(optional_array_query_param(mentioned_in)),
            project_id=none_as_unset(project_id),
            registry_id=none_as_unset(registry_id),
            schema_id=none_as_unset(schema_id),
            archive_reason=none_as_unset(archive_reason),
            mentions=none_as_unset(optional_array_query_param(mentions)),
            sort=none_as_unset(sort),
            ids=none_as_unset(optional_array_query_param(ids)),
            schema_fields=none_as_unset(schema_fields_query_param(schema_fields)),
            entity_registry_idsany_of=none_as_unset(optional_array_query_param(entity_registry_ids_any_of)),
            name_includes=none_as_unset(name_includes),
            namesany_of=none_as_unset(optional_array_query_param(names_any_of)),
            namesany_ofcase_sensitive=none_as_unset(optional_array_query_param(names_any_of_case_sensitive)),
            creator_ids=none_as_unset(optional_array_query_param(creator_ids)),
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
            author_idsany_of=none_as_unset(optional_array_query_param(author_idsany_of)),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(
        self,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        bases: Optional[str] = None,
        folder_id: Optional[str] = None,
        mentioned_in: Optional[List[str]] = None,
        project_id: Optional[str] = None,
        registry_id: Optional[str] = None,
        schema_id: Optional[str] = None,
        archive_reason: Optional[str] = None,
        mentions: Optional[List[str]] = None,
        ids: Optional[Iterable[str]] = None,
        entity_registry_ids_any_of: Optional[Iterable[str]] = None,
        name_includes: Optional[str] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        creator_ids: Optional[Iterable[str]] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
        sort: Optional[Union[ListRNASequencesSort]] = None,
        page_size: Optional[int] = None,
        author_idsany_of: Optional[Iterable[str]] = None,
    ) -> PageIterator[RnaSequence]:
        """
        List RNA sequences.

        See https://benchling.com/api/reference#/RNA%20Sequences/listRNASequences
        """
        if sort:
            sort = ListRNASequencesSort(sort)

        def api_call(next_token: NextToken) -> Response[RnaSequencesPaginatedList]:
            return self._rna_sequences_page(
                modified_at=modified_at,
                name=name,
                bases=bases,
                folder_id=folder_id,
                mentioned_in=mentioned_in,
                project_id=project_id,
                registry_id=registry_id,
                schema_id=schema_id,
                archive_reason=archive_reason,
                mentions=mentions,
                ids=ids,
                entity_registry_ids_any_of=entity_registry_ids_any_of,
                name_includes=name_includes,
                names_any_of=names_any_of,
                names_any_of_case_sensitive=names_any_of_case_sensitive,
                creator_ids=creator_ids,
                schema_fields=schema_fields,
                sort=sort,
                page_size=page_size,
                next_token=next_token,
                author_idsany_of=author_idsany_of,
            )

        def results_extractor(body: RnaSequencesPaginatedList) -> Optional[List[RnaSequence]]:
            return body.rna_sequences

        return PageIterator(api_call, results_extractor)

    @api_method
    def create(self, rna_sequence: RnaSequenceCreate) -> RnaSequence:
        """
        Create a RNA sequence.

        See https://benchling.com/api/reference#/RNA%20Sequences/createRNASequence
        """
        response = create_rna_sequence.sync_detailed(client=self.client, json_body=rna_sequence)
        return model_from_detailed(response)

    @api_method
    def update(self, rna_sequence_id: str, rna_sequence: RnaSequenceUpdate) -> RnaSequence:
        """
        Update a RNA sequence.

        See https://benchling.com/api/reference#/RNA%20Sequences/updateRNASequence
        """
        response = update_rna_sequence.sync_detailed(
            client=self.client, rna_sequence_id=rna_sequence_id, json_body=rna_sequence
        )
        return model_from_detailed(response)

    @api_method
    def archive(
        self, rna_sequence_ids: Iterable[str], reason: EntityArchiveReason
    ) -> RnaSequencesArchivalChange:
        """
        Archive RNA sequences.

        See https://benchling.com/api/reference#/RNA%20Sequences/archiveRNASequences
        """
        archive_request = RnaSequencesArchive(reason=reason, rna_sequence_ids=list(rna_sequence_ids))
        response = archive_rna_sequences.sync_detailed(client=self.client, json_body=archive_request)
        return model_from_detailed(response)

    @api_method
    def unarchive(self, rna_sequence_ids: Iterable[str]) -> RnaSequencesArchivalChange:
        """
        Unarchive RNA sequences.

        See https://benchling.com/api/reference#/RNA%20Sequences/unarchiveRNASequences
        """
        unarchive_request = RnaSequencesUnarchive(rna_sequence_ids=list(rna_sequence_ids))
        response = unarchive_rna_sequences.sync_detailed(client=self.client, json_body=unarchive_request)
        return model_from_detailed(response)

    @api_method
    def bulk_get(self, rna_sequence_ids: Iterable[str]) -> Optional[List[RnaSequence]]:
        """
        Bulk get RNA sequences.

        See https://benchling.com/api/reference#/RNA%20Sequences/bulkGetRNASequences
        """
        rna_sequence_id_string = ",".join(rna_sequence_ids)
        response = bulk_get_rna_sequences.sync_detailed(
            client=self.client, rna_sequence_ids=rna_sequence_id_string
        )
        rna_sequences_results = model_from_detailed(response)
        return rna_sequences_results.rna_sequences

    @api_method
    def bulk_create(self, rna_sequences: Iterable[RnaSequenceBulkCreate]) -> AsyncTaskLink:
        """
        Bulk create RNA sequences.

        See https://benchling.com/api/reference#/RNA%20Sequences/bulkCreateRNASequences
        """
        body = RnaSequencesBulkCreateRequest(list(rna_sequences))
        response = bulk_create_rna_sequences.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)

    @api_method
    def bulk_update(self, rna_sequences: Iterable[RnaSequenceBulkUpdate]) -> AsyncTaskLink:
        """
        Bulk update RNA sequences.

        See https://benchling.com/api/reference#/RNA%20Sequences/bulkUpdateRNASequences
        """
        body = RnaSequencesBulkUpdateRequest(list(rna_sequences))
        response = bulk_update_rna_sequences.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)

    @api_method
    def autofill_parts(self, rna_sequence_ids: Iterable[str]) -> AsyncTaskLink:
        """
        Autofill RNA sequence parts.

        See https://benchling.com/api/reference#/RNA%20Sequences/autofillRNASequenceParts
        """
        body = AutofillRnaSequences(rna_sequence_ids=list(rna_sequence_ids))
        response = autofill_rna_sequence_parts.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)

    @api_method
    def autofill_translations(self, rna_sequence_ids: Iterable[str]) -> AsyncTaskLink:
        """
        Autofill RNA sequence translations.

        See https://benchling.com/api/reference#/RNA%20Sequences/autofillRNASequenceTranslations
        """
        body = AutofillRnaSequences(rna_sequence_ids=list(rna_sequence_ids))
        response = autofill_rna_sequence_translations.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)

    @api_method
    def auto_annotate(self, auto_annotate: AutoAnnotateRnaSequences) -> AsyncTaskLink:
        """
        Auto-annotate RNA sequences with matching features from specified Feature Libraries.

        See https://benchling.com/api/reference#/RNA%20Sequences/autoAnnotateRnaSequences
        """
        response = auto_annotate_rna_sequences.sync_detailed(client=self.client, json_body=auto_annotate)
        return model_from_detailed(response)



import collections
from contextlib import ContextDecorator
from enum import Enum
from typing import Iterable, Union


class MissingExtrasError(Exception):
    """
    Missing extras error.

    Some SDK functionality requires optional extras packages are installed. Certain code paths
    may still be usable without them, so we ship all SDK features but will raise this error at
    runtime if a particular dependency is missing for code that requires it.
    """

    pass


class ExtrasPackage(Enum):
    """Names of optional extras packages in the SDK."""

    CRYPTOGRAPHY = "cryptography"
    PYTHON_JOSE = "python-jose"


class _required_packages_context(ContextDecorator):
    """
    Context Manager to run code with required packages.

    That will re-raise MissingExtrasError if it finds a missing module or import
    expected to be optionally installed as an extras package.
    """

    _required_packages: Iterable[ExtrasPackage]

    def __init__(
        self,
        required_packages: Union[ExtrasPackage, Iterable[ExtrasPackage]],
    ):
        """
        Initialize RequiredPackagesContext.

        :param required_packages: A list of enumerated packages expected for successful execution.
        """
        packages = (
            required_packages
            if isinstance(required_packages, collections.abc.Iterable)
            else [required_packages]
        )
        self._required_packages = packages

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if exc_type in (ImportError, ModuleNotFoundError):
            package_names = ", ".join([package.value for package in self._required_packages])
            raise MissingExtrasError(
                f"One or more missing optional packages must be installed to use this functionality. "
                f"Please install [{package_names}] as extras.",
                exc_value,
            )
        elif exc_value:
            raise exc_value



from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import date, datetime
import json
from typing import Any, Dict, Generic, List, Optional, TypeVar, Union

from benchling_sdk.models import (
    BooleanAppConfigItemType,
    DateAppConfigItemType,
    DatetimeAppConfigItemType,
    FieldType,
    FloatAppConfigItemType,
    IntegerAppConfigItemType,
    JsonAppConfigItemType,
    SecureTextAppConfigItemType,
    TextAppConfigItemType,
)

JsonType = Union[Dict[str, Any], List[Any], str, int, float, bool]
ScalarType = TypeVar("ScalarType", bool, date, datetime, float, int, JsonType, str)
# JsonType support requires object to be unioned. Currently we do it inline.
ScalarModelType = Union[bool, date, datetime, float, int, str]
# Enum values cannot be used in literals, so copy the strings
ScalarConfigItemType = Union[
    BooleanAppConfigItemType,
    DateAppConfigItemType,
    DatetimeAppConfigItemType,
    FloatAppConfigItemType,
    IntegerAppConfigItemType,
    JsonAppConfigItemType,
    SecureTextAppConfigItemType,
    TextAppConfigItemType,
]


class ScalarDefinition(ABC, Generic[ScalarType]):
    """
    Scalar definition.

    Map how ScalarConfigTypes values can be converted into corresponding Python types.
    """

    @classmethod
    def init(cls):
        """Init."""
        return cls()

    @classmethod
    @abstractmethod
    def from_str(cls, value: Optional[str]) -> Optional[ScalarType]:
        """
        From string.

        Given an optional string value of scalar configuration, produce an Optional instance of the
        specific ScalarType. For instance, converting str to int.

        Used when coercing Python types from string values in API responses.
        """
        pass


class BoolScalar(ScalarDefinition[bool]):
    """
    Bool Scalar.

    Turn a Boolean-like string value into bool. Any permutation of "true" - case insensitive - is interpreted
    as True. Any other non-empty string is False.
    """

    @classmethod
    def from_str(cls, value: Optional[str]) -> Optional[bool]:
        """Convert optional str to optional bool."""
        # Though the spec declares str, this is actually being sent in JSON as a real Boolean
        # So runtime check defensively
        if value is not None:
            if isinstance(value, bool):
                return value
            if value.lower() == "true":
                return True
            return False
        return None


class DateScalar(ScalarDefinition[date]):
    """
    Date Scalar.

    Turn an ISO formatted date like YYYY-MM-dd into a date.
    """

    @classmethod
    def from_str(cls, value: Optional[str]) -> Optional[date]:
        """Convert optional str to optional date."""
        return date.fromisoformat(value) if value is not None else None


class DateTimeScalar(ScalarDefinition[datetime]):
    """
    Date Time Scalar.

    Turn a date time string into datetime.
    """

    @classmethod
    def from_str(cls, value: Optional[str]) -> Optional[datetime]:
        """Convert optional str to optional datetime."""
        return datetime.strptime(value, cls.expected_format()) if value is not None else None

    @staticmethod
    def expected_format() -> str:
        """Return the expected date mask for parsing string to datetime."""
        return "%Y-%m-%d %H:%M:%S %p"


class IsoDateTimeScalar(ScalarDefinition[datetime]):
    """
    Iso Date Time Scalar.

    Turn a ISO 8601 date time string into datetime. Benchling fields use RFC 3339, unlike app config date times.
    """

    @classmethod
    def from_str(cls, value: Optional[str]) -> Optional[datetime]:
        """Convert optional str to optional datetime."""
        return datetime.fromisoformat(value) if value is not None else None


class FloatScalar(ScalarDefinition[float]):
    """
    Float Scalar.

    Turn a string into float. Assumes the string, if not empty, is a valid floating point.
    """

    @classmethod
    def from_str(cls, value: Optional[str]) -> Optional[float]:
        """Convert optional str to optional float."""
        return float(value) if value is not None else None


class IntScalar(ScalarDefinition[int]):
    """
    Int Scalar.

    Turn a string into int. Assumes the string, if not empty, is a valid integer.
    """

    @classmethod
    def from_str(cls, value: Optional[str]) -> Optional[int]:
        """Convert optional str to optional int."""
        return int(value) if value is not None else None


class JsonScalar(ScalarDefinition[JsonType]):
    """
    Json Scalar.

    Turn a string into JSON. Assumes the string is a valid JSON string.
    """

    @classmethod
    def from_str(cls, value: Optional[str]) -> Optional[JsonType]:
        """Convert optional str to optional JsonType."""
        return json.loads(value) if value is not None else None


class TextScalar(ScalarDefinition[str]):
    """
    Text Scalar.

    Text is already a string, so no conversion is performed.
    """

    @classmethod
    def from_str(cls, value: Optional[str]) -> Optional[str]:
        """Convert optional str to optional str. Implemented to appease ScalarDefinition contract."""
        return value


class SecureTextScalar(TextScalar):
    """
    Secure Text Scalar.

    Text is already a string, so no conversion is performed.
    """

    pass


def scalar_definition_from_field_type(field_type: FieldType) -> ScalarDefinition:
    """
    Scalar Definition From Field Type.

    Returns a ScalarDefinition for parsing a typed value given a field type.
    Assumes TextScalar (string) for any types not specifically enumerated.
    """
    if field_type == field_type.BOOLEAN:
        return BoolScalar.init()
    elif field_type == field_type.DATE:
        return DateScalar.init()
    elif field_type == field_type.DATETIME:
        return IsoDateTimeScalar.init()
    elif field_type == field_type.FLOAT:
        return FloatScalar.init()
    elif field_type == field_type.INTEGER:
        return IntScalar.init()
    elif field_type == field_type.JSON:
        return JsonScalar.init()
    elif field_type == field_type.TEXT:
        return TextScalar.init()
    # Assume all other types are str, which is safe for fields
    return TextScalar.init()


# Maps scalar types from the API into typed Python SDK scalar definitions
DEFAULT_SCALAR_DEFINITIONS: Dict[ScalarConfigItemType, ScalarDefinition] = {
    BooleanAppConfigItemType.BOOLEAN: BoolScalar.init(),
    DateAppConfigItemType.DATE: DateScalar.init(),
    DatetimeAppConfigItemType.DATETIME: DateTimeScalar.init(),
    FloatAppConfigItemType.FLOAT: FloatScalar.init(),
    IntegerAppConfigItemType.INTEGER: IntScalar.init(),
    JsonAppConfigItemType.JSON: JsonScalar.init(),
    SecureTextAppConfigItemType.SECURE_TEXT: SecureTextScalar.init(),
    TextAppConfigItemType.TEXT: TextScalar.init(),
}



from typing_extensions import Protocol

from benchling_sdk.helpers.package_helpers import _required_packages_context, ExtrasPackage


class _KeyUnwrappingFunction(Protocol):
    """
    Key Unwrapping Function.

    A function which accepts a ciphertext with a wrapped key and returns the data encrypting key as bytes.
    """

    def __call__(self, wrapped_key: bytes) -> bytes:
        pass


@_required_packages_context(ExtrasPackage.PYTHON_JOSE)
def _create_key_type(unwrapping_function: _KeyUnwrappingFunction):
    """
    Create an unwrapping Jose Key.

    Only implements the unwrap function. This key should not be used or distributed for any other purpose.
    """
    from jose.backends.base import Key

    # Create function to inject into the key to unwrap
    def _unwrap(wrapped_key: bytes) -> bytes:
        return unwrapping_function(wrapped_key)

    # We explicitly don't want to implement any other part of the key... just the unwrap
    class _RSAKeyDelegatedUnwrapping(Key):
        def sign(self, msg):
            raise NotImplementedError()

        def verify(self, msg, sig):
            raise NotImplementedError()

        def public_key(self):
            raise NotImplementedError()

        def to_pem(self):
            raise NotImplementedError()

        def to_dict(self):
            raise NotImplementedError()

        def encrypt(self, plain_text, aad=None):
            raise NotImplementedError()

        def decrypt(self, cipher_text, iv=None, aad=None, tag=None):
            raise NotImplementedError()

        def wrap_key(self, key_data):
            raise NotImplementedError()

        # Intentionally implemented
        def unwrap_key(self, wrapped_key):
            return _unwrap(wrapped_key)

    return _RSAKeyDelegatedUnwrapping



from typing import Iterable, List

from benchling_api_client.v2.stable.api.inventory import validate_barcodes
from benchling_api_client.v2.stable.client import Client

from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.retry_helpers import RetryStrategy
from benchling_sdk.models import BarcodesList, BarcodeValidationResult
from benchling_sdk.services.v2.base_service import BaseService


class InventoryService(BaseService):
    """
    Inventory.

    Manage inventory wide objects.

    See https://benchling.com/api/reference#/Inventory
    """

    def __init__(
        self,
        client: Client,
        retry_strategy: RetryStrategy,
    ):
        """
        Initialize inventory service.

        :param client: Underlying generated Client.
        :param retry_strategy: Retry strategy for failed HTTP calls
        """
        super().__init__(client, retry_strategy)

    @api_method
    def validate_barcodes(self, registry_id: str, barcodes: Iterable[str]) -> List[BarcodeValidationResult]:
        """
        Validate barcodes.

        See https://benchling.com/api/reference#/Inventory/validateBarcodes

        :param registry_id: ID of the registry to validate barcodes in.
        :param barcodes: The barcodes to validate
        :return: A list of bardcode validation results
        :rtype: List[BarcodeValidationResult]
        """
        barcodes_list = BarcodesList(barcodes=list(barcodes))
        response = validate_barcodes.sync_detailed(
            client=self.client, registry_id=registry_id, json_body=barcodes_list
        )
        results = model_from_detailed(response)
        return results.validation_results



from __future__ import annotations

from datetime import date, datetime
import json
import random
import string
from typing import Dict, List, Optional, Type, TypeVar, Union

from benchling_api_client.v2.beta.models.base_manifest_config import BaseManifestConfig
from benchling_api_client.v2.beta.models.benchling_app_manifest import BenchlingAppManifest
from benchling_api_client.v2.beta.models.dropdown_dependency import DropdownDependency
from benchling_api_client.v2.beta.models.entity_schema_dependency import EntitySchemaDependency
from benchling_api_client.v2.beta.models.field_definitions_manifest import FieldDefinitionsManifest
from benchling_api_client.v2.beta.models.manifest_array_config import ManifestArrayConfig
from benchling_api_client.v2.beta.models.manifest_scalar_config import ManifestScalarConfig
from benchling_api_client.v2.beta.models.resource_dependency import ResourceDependency
from benchling_api_client.v2.beta.models.scalar_config_types import ScalarConfigTypes
from benchling_api_client.v2.beta.models.schema_dependency import SchemaDependency
from benchling_api_client.v2.beta.models.schema_dependency_subtypes import (
    SchemaDependencySubtypes as SchemaDependencySubtypesBeta,
)
from benchling_api_client.v2.beta.models.workflow_task_schema_dependency import WorkflowTaskSchemaDependency
from benchling_api_client.v2.extensions import UnknownType
from benchling_api_client.v2.stable.types import UNSET, Unset
from typing_extensions import Protocol

from benchling_sdk.apps.config.decryption_provider import BaseDecryptionProvider
from benchling_sdk.apps.config.dependencies import (
    _supported_config_item,
    BaseDependencies,
    ConfigurationReference,
    DependencyLinkStore,
    StaticConfigProvider,
)
from benchling_sdk.apps.config.scalars import (
    BoolScalar,
    DateScalar,
    DateTimeScalar,
    FloatScalar,
    IntScalar,
    JsonScalar,
    JsonType,
    ScalarDefinition,
    ScalarType,
    SecureTextScalar,
    TextScalar,
)
from benchling_sdk.apps.helpers.config_helpers import (
    element_definition_from_dependency,
    field_definitions_from_dependency,
    options_from_dependency,
    subtype_from_entity_schema_dependency,
    workflow_task_schema_output_from_dependency,
)
from benchling_sdk.models import (
    AppConfigItem,
    ArrayElementAppConfigItem,
    ArrayElementAppConfigItemType,
    BooleanAppConfigItem,
    BooleanAppConfigItemType,
    DateAppConfigItem,
    DateAppConfigItemType,
    DatetimeAppConfigItem,
    DatetimeAppConfigItemType,
    EntitySchemaAppConfigItem,
    EntitySchemaAppConfigItemType,
    FieldAppConfigItem,
    FieldAppConfigItemType,
    FloatAppConfigItem,
    FloatAppConfigItemType,
    GenericApiIdentifiedAppConfigItem,
    GenericApiIdentifiedAppConfigItemType,
    IntegerAppConfigItem,
    IntegerAppConfigItemType,
    JsonAppConfigItem,
    JsonAppConfigItemType,
    LinkedAppConfigResourceSummary,
    SchemaDependencySubtypes,
    SecureTextAppConfigItem,
    SecureTextAppConfigItemType,
    TextAppConfigItem,
    TextAppConfigItemType,
)

ManifestDependencies = Union[
    DropdownDependency,
    EntitySchemaDependency,
    ManifestArrayConfig,
    ManifestScalarConfig,
    ResourceDependency,
    SchemaDependency,
    WorkflowTaskSchemaDependency,
    UnknownType,
]

D = TypeVar("D", bound=BaseDependencies)


class MockDecryptionFunction(Protocol):
    """Mock out a decryption function for use with secure text."""

    def __call__(self, ciphertext: str) -> str:
        """Return a string representing plaintext given input ciphertext."""


class MockDecryptionProvider(BaseDecryptionProvider):
    """
    Mock Decryption Provider.

    A generic class mocking a BaseDecryptionProvider. Can be passed a function to mock arbitrary decryption.

    It's recommended to extend this class or use a specific implementation instead of initializing an instance.
    """

    _mock_decryption_function: MockDecryptionFunction

    def __init__(self, mock_decryption_function: MockDecryptionFunction):
        """
        Init Mock Decryption Provider.

        Pass a function that returns desired mocked plaintext given ciphertext.
        """
        self._mock_decryption_function = mock_decryption_function

    def decrypt(self, ciphertext: str) -> str:
        """
        Decrypt.

        Invokes the mocked decryption function provided when instantiating the class to return a "decrypted" value.
        """
        return self._mock_decryption_function(ciphertext)


class MockBenchlingAppConfig:
    """
    Mock App Config.

    A helper class for easily mocking app config in various permutations.

    Easily mock all config items from a manifest model (which can be loaded from
    `benchling_sdk.apps.helpers.manifest_helpers.manifest_from_file()`.
    """

    _config_items: List[AppConfigItem]
    _decryption_provider: Optional[MockDecryptionProvider]

    def __init__(
        self, config_items: List[AppConfigItem], decryption_provider: Optional[MockDecryptionProvider]
    ):
        """
        Init Mock Benchling App Config.

        The class can be initialized by providing a list of AppConfigItem, but the recommended
        usage is to mock directly from a manifest, like `MockBenchlingAppConfig.from_manifest()`
        """
        self._config_items = config_items
        self._decryption_provider = decryption_provider

    @classmethod
    def from_manifest(
        cls, manifest: BenchlingAppManifest, decryption_provider: Optional[MockDecryptionProvider] = None
    ) -> MockBenchlingAppConfig:
        """
        From Manifest.

        Reads a manifest amd mocks out all dependencies.
        """
        config_items = mock_app_config_items_from_manifest(manifest)
        return cls(config_items, decryption_provider)

    def to_dependencies(self, target_dependencies: Type[D]) -> D:
        """
        To Dependencies.

        Convenience method for providing mocked app config to a target class extending BaseDependencies.
        """
        link_store = DependencyLinkStore(StaticConfigProvider(self.config_items))
        return target_dependencies.from_store(link_store, decryption_provider=self._decryption_provider)

    def with_replacement(self, replacement: AppConfigItem) -> MockBenchlingAppConfig:
        """
        With Replacement.

        Returns a new MockBenchlingAppConfig with the app config item at the specified path replaced.
        """
        # list() solves "List is invariant"
        replaced_app_config = replace_mocked_config_item_by_path(
            list(self.config_items), _config_path(replacement), replacement
        )
        return MockBenchlingAppConfig(replaced_app_config, self._decryption_provider)

    def with_replacements(self, replacements: List[AppConfigItem]) -> MockBenchlingAppConfig:
        """
        With Replacement.

        Returns a new MockBenchlingAppConfig with the app config item at the specified path replaced.
        """
        # list() solves "List is invariant"
        replaced_app_config: List[AppConfigItem] = list(self.config_items)
        for replacement in replacements:
            replaced_app_config = replace_mocked_config_item_by_path(
                list(replaced_app_config), _config_path(replacement), replacement
            )
        return MockBenchlingAppConfig(list(replaced_app_config), self._decryption_provider)

    @property
    def config_items(self) -> List[ConfigurationReference]:
        """List the config items in the mock app config, excluding any unknown types."""
        return [_supported_config_item(config_item) for config_item in self._config_items]

    @property
    def config_items_with_unknown(self) -> List[AppConfigItem]:
        """List the config items in the mock app config, including any unknown types."""
        return self._config_items


class MockDecryptionProviderStatic(MockDecryptionProvider):
    """
    Mock Decryption Provider Static.

    Always return the same "decrypted" value regardless of what ciphertext is passed.
    Useful if you only have a single secret value.
    """

    def __init__(self, decrypt_value: str):
        """
        Init Mock Decryption Provider Static.

        Supply the string to always be returned.
        """

        def decrypt(ciphertext: str) -> str:
            return decrypt_value

        super().__init__(decrypt)


class MockDecryptionProviderMapped(MockDecryptionProvider):
    """
    Mock Decryption Provider Mapped.

    Returns a "decrypted" value based on the input ciphertext.
    Useful if you have multiple secrets to mock simultaneously.
    """

    def __init__(self, decrypt_mapping: Dict[str, str]):
        """
        Init Mock Decryption Provider Mapped.

        Supply the dictionary mapping with ciphertext as keys and plaintext as values.
        Any ciphertext decrypted without a corresponding value will result in a KeyError.
        """

        def decrypt(ciphertext: str) -> str:
            return decrypt_mapping[ciphertext]

        super().__init__(decrypt)


def mock_app_config_items_from_manifest(manifest: BenchlingAppManifest) -> List[AppConfigItem]:
    """
    Mock Benchling App Config Items.

    This method accepts an app manifest model and creates mocked values for app the app config.

    The concrete mocked out values, such as API Ids and schema names are nonsensical and random,
    but are valid.

    Code should avoid relying on specific values or conventions (such as API prefixes). If
    specific dependency values need to be tested in isolation, the caller can selectively
    override the randomized values with replace_mocked_config_item_by_path().
    """
    root_config_items = [_mock_dependency(dependency) for dependency in manifest.configuration]
    return [config_item for sub_config_items in root_config_items for config_item in sub_config_items]


def replace_mocked_config_item_by_path(
    original_config: List[AppConfigItem], replacement_path: List[str], replacement_item: AppConfigItem
) -> List[AppConfigItem]:
    """Return an updated list of app config items with a specific config item replaced with the provided mock."""
    replaced_config = [config for config in original_config if _config_path(config) != replacement_path]
    replaced_config.append(replacement_item)
    return replaced_config


def mock_bool_app_config_item(path: List[str], value: Optional[bool]) -> BooleanAppConfigItem:
    """Mock a bool app config item with a path and specified value."""
    return BooleanAppConfigItem(
        path=path,
        value=value,
        type=BooleanAppConfigItemType.BOOLEAN,
        id=_random_string("aci_"),
    )


def mock_date_app_config_item(path: List[str], value: Optional[date]) -> DateAppConfigItem:
    """Mock a date app config item with a path and specified value."""
    return DateAppConfigItem(
        path=path,
        value=str(value) if value is not None else None,
        type=DateAppConfigItemType.DATE,
        id=_random_string("aci_"),
    )


def mock_datetime_app_config_item(path: List[str], value: Optional[datetime]) -> DatetimeAppConfigItem:
    """Mock a datetime app config item with a path and specified value."""
    return DatetimeAppConfigItem(
        path=path,
        value=value.strftime(DateTimeScalar.expected_format()) if isinstance(value, datetime) else value,
        type=DatetimeAppConfigItemType.DATETIME,
        id=_random_string("aci_"),
    )


def mock_float_app_config_item(path: List[str], value: Optional[float]) -> FloatAppConfigItem:
    """Mock a float app config item with a path and specified value."""
    return FloatAppConfigItem(
        path=path,
        value=value,
        type=FloatAppConfigItemType.FLOAT,
        id=_random_string("aci_"),
    )


def mock_int_app_config_item(path: List[str], value: Optional[int]) -> IntegerAppConfigItem:
    """Mock an int app config item with a path and specified value."""
    return IntegerAppConfigItem(
        path=path,
        value=value,
        type=IntegerAppConfigItemType.INTEGER,
        id=_random_string("aci_"),
    )


def mock_json_app_config_item(path: List[str], value: Optional[JsonType]) -> JsonAppConfigItem:
    """Mock an int app config item with a path and specified value."""
    return JsonAppConfigItem(
        path=path,
        value=json.dumps(value) if value is not None else None,
        type=JsonAppConfigItemType.JSON,
        id=_random_string("aci_"),
    )


def mock_secure_text_app_config_item(path: List[str], value: Optional[str]) -> SecureTextAppConfigItem:
    """Mock a secure text app config item with a path and specified value."""
    return SecureTextAppConfigItem(
        path=path,
        value=value,
        type=SecureTextAppConfigItemType.SECURE_TEXT,
        id=_random_string("aci_"),
    )


def mock_text_app_config_item(path: List[str], value: Optional[str]) -> TextAppConfigItem:
    """Mock a text app config item with a path and specified value."""
    return TextAppConfigItem(
        path=path,
        value=value,
        type=TextAppConfigItemType.TEXT,
        id=_random_string("aci_"),
    )


def _mock_dependency(
    dependency: ManifestDependencies,
    parent_path: List[str] = list(),
) -> List[AppConfigItem]:
    """Mock a dependency from its manifest definition."""
    # MyPy has trouble inferring lists with [config_item] + sub_items so use the syntax like:
    # [*[config_item], *sub_items]
    # See https://github.com/python/mypy/issues/3933#issuecomment-808739063
    if isinstance(dependency, DropdownDependency):
        linked_resource_id = _random_string("val_")
        config_item = GenericApiIdentifiedAppConfigItem(
            id=_random_string("aci_"),
            path=parent_path + [dependency.name],
            type=GenericApiIdentifiedAppConfigItemType.DROPDOWN,
            value=_random_string("val_"),
            linked_resource=_mock_linked_resource(linked_resource_id),
        )
        sub_items = [
            _mock_subdependency(subdependency, dependency, parent_path=parent_path)
            for subdependency in options_from_dependency(dependency)
        ]
        return [*[config_item], *sub_items]
    elif isinstance(dependency, EntitySchemaDependency):
        linked_resource_id = _random_string("val_")
        subtype = subtype_from_entity_schema_dependency(dependency)
        optional_subtype: Union[SchemaDependencySubtypes, Unset] = (
            _convert_entity_subtype(subtype) if subtype is not None else UNSET
        )
        entity_item = EntitySchemaAppConfigItem(
            id=_random_string("aci_"),
            path=parent_path + [dependency.name],
            type=EntitySchemaAppConfigItemType.ENTITY_SCHEMA,
            subtype=optional_subtype,
            value=_random_string("val_"),
            linked_resource=_mock_linked_resource(linked_resource_id),
        )
        sub_items = [
            _mock_subdependency(subdependency, dependency, parent_path=parent_path)
            for subdependency in field_definitions_from_dependency(dependency)
        ]
        return [*[entity_item], *sub_items]
    elif isinstance(dependency, SchemaDependency):
        linked_resource_id = _random_string("val_")
        config_item = GenericApiIdentifiedAppConfigItem(
            id=_random_string("aci_"),
            path=parent_path + [dependency.name],
            type=GenericApiIdentifiedAppConfigItemType(dependency.type),
            value=_random_string("val_"),
            linked_resource=_mock_linked_resource(linked_resource_id),
        )
        sub_items = [
            _mock_subdependency(subdependency, dependency, parent_path=parent_path)
            for subdependency in field_definitions_from_dependency(dependency)
        ]
        return [*[config_item], *sub_items]
    elif isinstance(dependency, WorkflowTaskSchemaDependency):
        linked_resource_id = _random_string("val_")
        config_item = GenericApiIdentifiedAppConfigItem(
            id=_random_string("aci_"),
            path=parent_path + [dependency.name],
            type=GenericApiIdentifiedAppConfigItemType.WORKFLOW_TASK_SCHEMA,
            value=linked_resource_id,
            linked_resource=_mock_linked_resource(linked_resource_id),
        )
        sub_items = [
            _mock_subdependency(subdependency, dependency, parent_path=parent_path)
            for subdependency in field_definitions_from_dependency(dependency)
        ]
        workflow_task_output = workflow_task_schema_output_from_dependency(dependency)
        if workflow_task_output:
            output_fields = field_definitions_from_dependency(workflow_task_output)
            output_items = [
                _mock_workflow_output_subdependency(subdependency, dependency, parent_path=parent_path)
                for subdependency in output_fields
            ]
            sub_items += output_items
        return [*[config_item], *sub_items]
    elif isinstance(dependency, ManifestScalarConfig):
        return [_mock_scalar_dependency(dependency, parent_path=parent_path)]
    elif isinstance(dependency, ManifestArrayConfig):
        return _mock_array_dependency(dependency, parent_path=parent_path)
    elif isinstance(dependency, UnknownType):
        return [UnknownType(value="Unknown")]
    else:
        linked_resource_id = _random_string("val_")
        return [
            GenericApiIdentifiedAppConfigItem(
                id=_random_string("aci_"),
                path=parent_path + [dependency.name],
                type=GenericApiIdentifiedAppConfigItemType(dependency.type),
                value=linked_resource_id,
                linked_resource=_mock_linked_resource(linked_resource_id),
            )
        ]


def _convert_entity_subtype(manifest_subtype: SchemaDependencySubtypesBeta) -> SchemaDependencySubtypes:
    # Manifest types and config types should be equivalent but are technically different Enums
    source_value: str = manifest_subtype.value
    return SchemaDependencySubtypes(source_value)


def _mock_scalar_dependency(
    dependency: ManifestScalarConfig, parent_path: List[str] = list()
) -> AppConfigItem:
    if dependency.type == ScalarConfigTypes.BOOLEAN:
        bool_value = _mock_scalar_value_with_conversion(dependency, BoolScalar)
        bool_config = mock_bool_app_config_item([dependency.name], bool_value)
        return _append_config_item_path(bool_config, parent_path)
    elif dependency.type == ScalarConfigTypes.DATE:
        date_value = _mock_scalar_value_with_conversion(dependency, DateScalar)
        date_config = mock_date_app_config_item([dependency.name], date_value)
        return _append_config_item_path(date_config, parent_path)
    elif dependency.type == ScalarConfigTypes.DATETIME:
        datetime_value = _mock_scalar_value_with_conversion(dependency, DateTimeScalar)
        datetime_config = mock_datetime_app_config_item([dependency.name], datetime_value)
        return _append_config_item_path(datetime_config, parent_path)
    elif dependency.type == ScalarConfigTypes.FLOAT:
        float_value = _mock_scalar_value_with_conversion(dependency, FloatScalar)
        float_config = mock_float_app_config_item([dependency.name], float_value)
        return _append_config_item_path(float_config, parent_path)
    elif dependency.type == ScalarConfigTypes.INTEGER:
        int_value = _mock_scalar_value_with_conversion(dependency, IntScalar)
        int_config = mock_int_app_config_item([dependency.name], int_value)
        return _append_config_item_path(int_config, parent_path)
    elif dependency.type == ScalarConfigTypes.JSON:
        json_value = _mock_scalar_value_with_conversion(dependency, JsonScalar)
        json_config = mock_json_app_config_item([dependency.name], json_value)
        return _append_config_item_path(json_config, parent_path)
    elif dependency.type == ScalarConfigTypes.SECURE_TEXT:
        secure_text_value = _mock_scalar_value_with_conversion(dependency, SecureTextScalar)
        secure_text_config = mock_secure_text_app_config_item([dependency.name], secure_text_value)
        return _append_config_item_path(secure_text_config, parent_path)
    else:
        text_value = _mock_scalar_value_with_conversion(dependency, TextScalar)
        text_config = mock_text_app_config_item([dependency.name], text_value)
        return _append_config_item_path(text_config, parent_path)


def _append_config_item_path(config_item: AppConfigItem, parent_path: List[str]) -> AppConfigItem:
    if isinstance(config_item, UnknownType):
        return config_item
    config_item.path = parent_path + config_item.path
    return config_item


def _mock_array_dependency(
    dependency: ManifestArrayConfig, parent_path: List[str] = list(), rows: int = 1
) -> List[AppConfigItem]:
    config_rows = []
    for i in range(rows):
        row = _mock_array_row(dependency, parent_path=parent_path)
        elements = element_definition_from_dependency(dependency)
        element_configs = [_mock_dependency(element, row.path) for element in elements]
        flattened_configs = [element for sublist in element_configs for element in sublist]
        config_rows.append(row)
        config_rows.extend(flattened_configs)
    return config_rows


def _mock_array_row(dependency: ManifestArrayConfig, parent_path: List[str] = list()):
    row_name = _random_string("Row ")
    return ArrayElementAppConfigItem(
        id=_random_string("aci_"),
        path=parent_path + [dependency.name, row_name],
        type=ArrayElementAppConfigItemType.ARRAY_ELEMENT,
        value=row_name,
    )


def _mock_scalar_value_with_conversion(
    dependency: ManifestScalarConfig, scalar_definition_type: Type[ScalarDefinition[ScalarType]]
) -> Optional[ScalarType]:
    mocked_value_str = _mock_scalar_value(dependency.type)
    return scalar_definition_type.from_str(mocked_value_str)


def _mock_subdependency(
    subdependency: Union[BaseManifestConfig, FieldDefinitionsManifest],
    parent_config,
    parent_path: List[str] = list(),
) -> AppConfigItem:
    if isinstance(parent_config, DropdownDependency):
        linked_resource_id = _random_string("opt_")
        return GenericApiIdentifiedAppConfigItem(
            id=_random_string("aci_"),
            path=parent_path + [parent_config.name, subdependency.name],
            type=GenericApiIdentifiedAppConfigItemType.DROPDOWN_OPTION,
            value=linked_resource_id,
            linked_resource=_mock_linked_resource(linked_resource_id),
        )
    elif isinstance(parent_config, (EntitySchemaDependency, SchemaDependency, WorkflowTaskSchemaDependency)):
        path = [parent_config.name, subdependency.name]
        linked_resource_id = _random_string("tsf_")
        app_config = FieldAppConfigItem(
            id=_random_string("aci_"),
            path=parent_path + path,
            type=FieldAppConfigItemType.FIELD,
            value=linked_resource_id,
            linked_resource=_mock_linked_resource(linked_resource_id),
        )
        return app_config
    raise RuntimeError(f"Can't mock unsupported dependency ({subdependency})")


def _mock_workflow_output_subdependency(
    subdependency: Union[BaseManifestConfig, FieldDefinitionsManifest],
    parent_config,
    parent_path: List[str] = list(),
) -> AppConfigItem:
    linked_resource_id = _random_string("tsf_")
    app_config = FieldAppConfigItem(
        id=_random_string("aci_"),
        path=parent_path + [parent_config.name, "output", subdependency.name],
        type=FieldAppConfigItemType.FIELD,
        value=linked_resource_id,
        linked_resource=_mock_linked_resource(linked_resource_id),
    )
    return app_config


def _mock_linked_resource(id: str, name: Optional[str] = None) -> LinkedAppConfigResourceSummary:
    return LinkedAppConfigResourceSummary(id=id, name=name if name else _random_string("Resource Name"))


def _mock_scalar_value(scalar_type: ScalarConfigTypes) -> str:
    """Mock a scalar config value from its manifest definition."""
    if scalar_type == scalar_type.BOOLEAN:
        return "true"
    elif scalar_type == scalar_type.DATE:
        return date.today().strftime("%Y-%m-%d")
    elif scalar_type == scalar_type.DATETIME:
        return datetime.now().strftime(DateTimeScalar.expected_format())
    elif scalar_type == scalar_type.FLOAT:
        return str(random.random())
    elif scalar_type == scalar_type.INTEGER:
        return str(random.randint(-1000, 1000))
    elif scalar_type == scalar_type.JSON:
        return json.dumps(
            {_random_string(): [_random_string(), _random_string()], _random_string(): random.random()}
        )
    return _random_string()


def _random_string(prefix: str = "", random_length: int = 20) -> str:
    """Generate a randomized string up to a specified length with an optional prefix."""
    delimited_prefix = f"{prefix}-" if prefix else ""
    return f"{delimited_prefix}{''.join(random.choice(string.ascii_letters) for i in range(random_length))}"


def _config_path(config_item: AppConfigItem) -> List[str]:
    if isinstance(config_item, UnknownType):
        return config_item.value["path"]
    return config_item.path



from benchling_api_client.v2.beta.api.dna_sequences import bulk_upsert_dna_sequences, upsert_dna_sequence
from benchling_api_client.v2.beta.models.dna_sequence_upsert_request import DnaSequenceUpsertRequest
from benchling_api_client.v2.beta.models.dna_sequences_bulk_upsert_request import (
    DnaSequencesBulkUpsertRequest,
)

from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.models import AsyncTaskLink, DnaSequence
from benchling_sdk.services.v2.base_service import BaseService


class V2BetaDnaSequenceService(BaseService):
    """
    V2-Beta DNA Sequences.

    DNA sequences are the bread and butter of the Benchling Molecular Biology suite. On Benchling, these are
    comprised of a string of nucleotides and collections of other attributes, such as annotations and primers.

    See https://benchling.com/api/v2-beta/reference#/DNA%20Sequences
    """

    @api_method
    def upsert(self, entity_registry_id: str, dna_sequence: DnaSequenceUpsertRequest) -> DnaSequence:
        """
        Create or modify a DNA sequence.

        See https://benchling.com/api/v2-beta/reference#/DNA%20Sequences/upsertDNASequence
        """
        response = upsert_dna_sequence.sync_detailed(
            client=self.client, entity_registry_id=entity_registry_id, json_body=dna_sequence
        )
        return model_from_detailed(response)

    @api_method
    def bulk_upsert(self, body: DnaSequencesBulkUpsertRequest) -> AsyncTaskLink:
        """
        Bulk create or update DNA sequences.

        See https://benchling.com/api/v2-beta/reference#/DNA%20Sequences/bulkUpsertDnaSequences
        """
        response = bulk_upsert_dna_sequences.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)



from typing import Any, Dict, Iterable, List, Optional, Union

from benchling_api_client.v2.stable.api.dna_oligos import (
    archive_dna_oligos,
    bulk_create_dna_oligos,
    bulk_update_dna_oligos,
    create_dna_oligo,
    get_dna_oligo,
    list_dna_oligos,
    unarchive_dna_oligos,
    update_dna_oligo,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.constants import _translate_to_string_enum
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import (
    none_as_unset,
    optional_array_query_param,
    schema_fields_query_param,
)
from benchling_sdk.models import (
    AsyncTaskLink,
    DnaOligo,
    DnaOligoBulkUpdate,
    DnaOligoCreate,
    DnaOligosArchivalChange,
    DnaOligosArchive,
    DnaOligosBulkCreateRequest,
    DnaOligosBulkUpdateRequest,
    DnaOligosPaginatedList,
    DnaOligosUnarchive,
    DnaOligoUpdate,
    EntityArchiveReason,
    ListDNAOligosSort,
)
from benchling_sdk.services.v2.base_service import BaseService


class DnaOligoService(BaseService):
    """
    DNA Oligos.

    DNA Oligos are short linear DNA sequences that can be attached as primers to full DNA sequences. Just like other
    entities, they support schemas, tags, and aliases.
    See https://benchling.com/api/reference#/DNA%20Oligos
    """

    @api_method
    def get_by_id(self, oligo_id: str) -> DnaOligo:
        """
        Get a DNA Oligo by ID.

        See https://benchling.com/api/reference#/DNA%20Oligos/getDNAOligo
        """
        response = get_dna_oligo.sync_detailed(client=self.client, oligo_id=oligo_id)
        return model_from_detailed(response)

    @api_method
    def _dna_oligos_page(
        self,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        bases: Optional[str] = None,
        folder_id: Optional[str] = None,
        mentioned_in: Optional[List[str]] = None,
        project_id: Optional[str] = None,
        registry_id: Optional[str] = None,
        schema_id: Optional[str] = None,
        archive_reason: Optional[str] = None,
        mentions: Optional[List[str]] = None,
        sort: Optional[ListDNAOligosSort] = None,
        ids: Optional[Iterable[str]] = None,
        entity_registry_ids_any_of: Optional[Iterable[str]] = None,
        name_includes: Optional[str] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
        creator_ids: Optional[Iterable[str]] = None,
        page_size: Optional[int] = None,
        next_token: NextToken = None,
        author_idsany_of: Optional[Iterable[str]] = None,
    ) -> Response[DnaOligosPaginatedList]:
        response = list_dna_oligos.sync_detailed(
            client=self.client,
            modified_at=none_as_unset(modified_at),
            name=none_as_unset(name),
            bases=none_as_unset(bases),
            folder_id=none_as_unset(folder_id),
            mentioned_in=none_as_unset(optional_array_query_param(mentioned_in)),
            project_id=none_as_unset(project_id),
            registry_id=none_as_unset(registry_id),
            schema_id=none_as_unset(schema_id),
            archive_reason=none_as_unset(archive_reason),
            mentions=none_as_unset(optional_array_query_param(mentions)),
            sort=none_as_unset(sort),
            ids=none_as_unset(optional_array_query_param(ids)),
            entity_registry_idsany_of=none_as_unset(optional_array_query_param(entity_registry_ids_any_of)),
            name_includes=none_as_unset(name_includes),
            namesany_of=none_as_unset(optional_array_query_param(names_any_of)),
            namesany_ofcase_sensitive=none_as_unset(optional_array_query_param(names_any_of_case_sensitive)),
            schema_fields=none_as_unset(schema_fields_query_param(schema_fields)),
            creator_ids=none_as_unset(optional_array_query_param(creator_ids)),
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
            author_idsany_of=none_as_unset(optional_array_query_param(author_idsany_of)),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(
        self,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        bases: Optional[str] = None,
        folder_id: Optional[str] = None,
        mentioned_in: Optional[List[str]] = None,
        project_id: Optional[str] = None,
        registry_id: Optional[str] = None,
        schema_id: Optional[str] = None,
        archive_reason: Optional[str] = None,
        mentions: Optional[List[str]] = None,
        ids: Optional[Iterable[str]] = None,
        entity_registry_ids_any_of: Optional[Iterable[str]] = None,
        name_includes: Optional[str] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
        creator_ids: Optional[Iterable[str]] = None,
        sort: Optional[Union[str, ListDNAOligosSort]] = None,
        page_size: Optional[int] = None,
        author_idsany_of: Optional[Iterable[str]] = None,
    ) -> PageIterator[DnaOligo]:
        """
        List DNA Oligos.

        See https://benchling.com/api/reference#/DNA%20Oligos/listDNAOligos
        """

        def api_call(next_token: NextToken) -> Response[DnaOligosPaginatedList]:
            return self._dna_oligos_page(
                modified_at=modified_at,
                name=name,
                bases=bases,
                folder_id=folder_id,
                mentioned_in=mentioned_in,
                project_id=project_id,
                registry_id=registry_id,
                schema_id=schema_id,
                archive_reason=archive_reason,
                mentions=mentions,
                ids=ids,
                entity_registry_ids_any_of=entity_registry_ids_any_of,
                name_includes=name_includes,
                names_any_of=names_any_of,
                names_any_of_case_sensitive=names_any_of_case_sensitive,
                schema_fields=schema_fields,
                creator_ids=creator_ids,
                sort=_translate_to_string_enum(ListDNAOligosSort, sort),
                page_size=page_size,
                next_token=next_token,
                author_idsany_of=author_idsany_of,
            )

        def results_extractor(body: DnaOligosPaginatedList) -> Optional[List[DnaOligo]]:
            return body.dna_oligos

        return PageIterator(api_call, results_extractor)

    @api_method
    def create(self, dna_oligo: DnaOligoCreate) -> DnaOligo:
        """
        Create a DNA Oligo.

        See https://benchling.com/api/reference#/DNA%20Oligos/createDNAOligo
        """
        response = create_dna_oligo.sync_detailed(client=self.client, json_body=dna_oligo)
        return model_from_detailed(response)

    @api_method
    def update(self, oligo_id: str, dna_oligo: DnaOligoUpdate) -> DnaOligo:
        """
        Update a DNA Oligo.

        See https://benchling.com/api/reference#/DNA%20Oligos/updateDNAOligo
        """
        response = update_dna_oligo.sync_detailed(client=self.client, oligo_id=oligo_id, json_body=dna_oligo)
        return model_from_detailed(response)

    @api_method
    def archive(self, dna_oligo_ids: Iterable[str], reason: EntityArchiveReason) -> DnaOligosArchivalChange:
        """
        Archive DNA Oligos.

        See https://benchling.com/api/reference#/DNA%20Oligos/archiveDNAOligos
        """
        archive_request = DnaOligosArchive(reason=reason, dna_oligo_ids=list(dna_oligo_ids))
        response = archive_dna_oligos.sync_detailed(client=self.client, json_body=archive_request)
        return model_from_detailed(response)

    @api_method
    def unarchive(self, dna_oligo_ids: Iterable[str]) -> DnaOligosArchivalChange:
        """
        Unarchive DNA Oligos.

        See https://benchling.com/api/reference#/DNA%20Oligos/unarchiveDNAOligos
        """
        unarchive_request = DnaOligosUnarchive(dna_oligo_ids=list(dna_oligo_ids))
        response = unarchive_dna_oligos.sync_detailed(client=self.client, json_body=unarchive_request)
        return model_from_detailed(response)

    @api_method
    def bulk_create(self, dna_oligos: Iterable[DnaOligoCreate]) -> AsyncTaskLink:
        """
        Bulk create DNA Oligos.

        See https://benchling.com/api/reference#/DNA%20Oligos/bulkCreateDNAOligos
        """
        body = DnaOligosBulkCreateRequest(list(dna_oligos))
        response = bulk_create_dna_oligos.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)

    @api_method
    def bulk_update(self, dna_oligos: Iterable[DnaOligoBulkUpdate]) -> AsyncTaskLink:
        """
        Bulk update DNA oligos.

        See https://benchling.com/api/reference#/DNA%20Oligos/bulkUpdateDNAOligos
        """
        body = DnaOligosBulkUpdateRequest(list(dna_oligos))
        response = bulk_update_dna_oligos.sync_detailed(client=self.client, json_body=body)
        return model_from_detailed(response)
